from typing import Optional

from great_expectations.core import (
    ExpectationValidationResult,
)
from great_expectations.execution_engine import PandasExecutionEngine
from great_expectations.expectations.expectation import (
    ColumnMapExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.expectations.metrics import (
    ColumnMapMetricProvider,
    column_condition_partial,
)
from great_expectations.render import RenderedStringTemplateContent
from great_expectations.render.components import LegacyRendererType
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.util import (
    num_to_str,
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)


class ColumnValuesEqualThree(ColumnMapMetricProvider):
    condition_metric_name = "column_values.equal_three"

    @column_condition_partial(engine=PandasExecutionEngine)
    def _pandas(cls, column, **kwargs):
        return column == 3


class ExpectColumnValuesToEqualThree(ColumnMapExpectation):
    map_metric = "column_values.equal_three"
    success_keys = ("mostly",)

    def validate_configuration(self, configuration) -> None:
        pass  # no-op to make test setup easier


class ExpectColumnValuesToEqualThree__SecondIteration(ExpectColumnValuesToEqualThree):
    """Expect values in this column to equal the number three."""

    examples = [
        {
            "dataset_name": "mostly_threes_second_iteration",
            "data": {
                "mostly_threes": [3, 3, 3, 3, 3, 3, 2, -1, None, None],
            },
            "tests": [
                {
                    "title": "positive_test_with_mostly",
                    "exact_match_out": False,
                    "in": {"column": "mostly_threes", "mostly": 0.6},
                    "include_in_gallery": True,
                    "out": {
                        "success": True,
                        "unexpected_index_list": [6, 7],
                        "unexpected_list": [2, -1],
                    },
                },
                {
                    "title": "negative_test_with_mostly",
                    "exact_match_out": False,
                    "in": {"column": "mostly_threes", "mostly": 0.9},
                    "include_in_gallery": False,
                    "out": {
                        "success": False,
                        "unexpected_index_list": [6, 7],
                        "unexpected_list": [2, -1],
                    },
                },
                {
                    "title": "other_negative_test_with_mostly",
                    "exact_match_out": False,
                    "in": {"column": "mostly_threes", "mostly": 0.9},
                    # "include_in_gallery": False, #This key is omitted, so the example shouldn't show up in the gallery  # noqa: E501 # FIXME CoP
                    "out": {
                        "success": False,
                        "unexpected_index_list": [6, 7],
                        "unexpected_list": [2, -1],
                    },
                },
            ],
        }
    ]

    library_metadata = {
        "maturity": "EXPERIMENTAL",
        "tags": ["tag", "other_tag"],
        "contributors": [
            "@abegong",
        ],
    }


class ExpectColumnValuesToEqualThree__ThirdIteration(
    ExpectColumnValuesToEqualThree__SecondIteration
):
    @classmethod
    @renderer(renderer_type="renderer.question")
    def _question_renderer(cls, configuration, result=None, runtime_configuration=None):
        column = configuration.kwargs.get("column")
        mostly = configuration.kwargs.get("mostly")

        if mostly:
            return f'Do at least {mostly * 100}% of values in column "{column}" equal 3?'
        else:
            return f'Do all the values in column "{column}" equal 3?'

    @classmethod
    @renderer(renderer_type="renderer.answer")
    def _answer_renderer(cls, configuration=None, result=None, runtime_configuration=None):
        column = result.expectation_config.kwargs.get("column")
        mostly = result.expectation_config.kwargs.get("mostly")

        if mostly:
            if result.success:
                return f'At least {mostly * 100}% of values in column "{column}" equal 3.'
            else:
                return f'Less than {mostly * 100}% of values in column "{column}" equal 3.'
        else:  # noqa: PLR5501 # FIXME CoP
            if result.success:
                return f'All of the values in column "{column}" equal 3.'
            else:
                return f'Not all of the values in column "{column}" equal 3.'

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        include_column_name = runtime_configuration.get("include_column_name") is not False
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration.kwargs,
            ["column", "regex", "mostly", "row_condition", "condition_parser"],
        )

        template_str = "values must be equal to 3"
        if params["mostly"] is not None:
            params["mostly_pct"] = num_to_str(params["mostly"] * 100, no_scientific=True)
            # params["mostly_pct"] = "{:.14f}".format(params["mostly"]*100).rstrip("0").rstrip(".")
            template_str += ", at least $mostly_pct % of the time."
        else:
            template_str += "."

        if include_column_name:
            template_str = "$column " + template_str

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = conditional_template_str + ", then " + template_str
            params.update(conditional_params)

        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]


class ExpectColumnValuesToEqualThree__BrokenIteration(
    ExpectColumnValuesToEqualThree__SecondIteration
):
    examples = [
        {
            "dataset_name": "mostly_threes_third_broken_iteration",
            "data": {
                "mostly_threes": [3, 3, 3, 3, 3, 3, 2, -1, None, None],
                "broken_column": [3, 3, 3, 3, 3, 3, 2, -1, "b", "b"],
            },
            "tests": [
                {
                    "title": "positive_test_with_mostly",
                    "exact_match_out": False,
                    "in": {"column": "mostly_threes", "mostly": 0.6},
                    "include_in_gallery": True,
                    "out": {
                        "success": True,
                        "unexpected_index_list": [6, 7],
                        "unexpected_list": [2, -1],
                    },
                },
                {
                    "title": "negative_test_with_mostly",
                    "exact_match_out": False,
                    "in": {"column": "mostly_threes", "mostly": 0.9},
                    "include_in_gallery": False,
                    "out": {
                        "success": False,
                        "unexpected_index_list": [6, 7],
                        "unexpected_list": [2, -1],
                    },
                },
                {
                    "title": "other_negative_test_with_mostly",
                    "exact_match_out": False,
                    "in": {"column": "mostly_threes", "mostly": 0.9},
                    # "include_in_gallery": False, #This key is omitted, so the example shouldn't show up in the gallery  # noqa: E501 # FIXME CoP
                    "out": {
                        "success": False,
                        "unexpected_index_list": [6, 7],
                        "unexpected_list": [2, -1],
                    },
                },
                {
                    "title": "test_that_will_error_out",
                    "exact_match_out": False,
                    "in": {"column": "column_that_doesnt_exist"},
                    "include_in_gallery": True,
                    "out": {
                        "success": True,
                        "unexpected_index_list": [6, 7],
                        "unexpected_list": [2, -1],
                    },
                },
                {
                    "title": "another_test_that_will_error_out",
                    "exact_match_out": False,
                    "in": {"column": "broken_column"},
                    "include_in_gallery": True,
                    "out": {
                        "success": True,
                        "unexpected_index_list": [6, 7],
                        "unexpected_list": [2, -1],
                    },
                },
            ],
        }
    ]



from __future__ import annotations

import logging
import smtplib
from contextlib import contextmanager
from datetime import datetime, timezone
from types import ModuleType
from typing import TYPE_CHECKING, Iterator, Literal
from unittest import mock

import pytest
import requests
from requests import Session

import great_expectations as gx
from great_expectations.checkpoint.actions import (
    ActionContext,
    APINotificationAction,
    EmailAction,
    MicrosoftTeamsNotificationAction,
    OpsgenieAlertAction,
    PagerdutyAlertAction,
    SlackNotificationAction,
    SNSNotificationAction,
    UpdateDataDocsAction,
    ValidationAction,
)
from great_expectations.checkpoint.checkpoint import Checkpoint, CheckpointResult
from great_expectations.core.batch import IDDict, LegacyBatchDefinition
from great_expectations.core.expectation_validation_result import (
    ExpectationSuiteValidationResult,
)
from great_expectations.core.run_identifier import RunIdentifier
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.data_context.abstract_data_context import (
    AbstractDataContext,
)
from great_expectations.data_context.data_context.cloud_data_context import CloudDataContext
from great_expectations.data_context.data_context.context_factory import (
    project_manager,
    set_context,
)
from great_expectations.data_context.types.resource_identifiers import (
    ExpectationSuiteIdentifier,
    GXCloudIdentifier,
    ValidationResultIdentifier,
)
from great_expectations.exceptions.exceptions import ValidationActionAlreadyRegisteredError
from great_expectations.util import is_library_loadable

if TYPE_CHECKING:
    from pytest_mock import MockerFixture
    from typing_extensions import Never

logger = logging.getLogger(__name__)


# Global constants to be referenced in both tests and fixtures
SUITE_A: str = "suite_a"
SUITE_B: str = "suite_b"
BATCH_ID_A: str = "my_datasource-my_first_asset"
BATCH_ID_B: str = "my_datasource-my_second_asset"
utc_datetime = datetime.fromisoformat("2024-04-01T20:51:18.077262").replace(tzinfo=timezone.utc)


@pytest.fixture
def mocked_posthog(mocker: MockerFixture):
    yield mocker.patch("posthog.capture")


@pytest.fixture
def checkpoint_result(mocker: MockerFixture):
    utc_datetime = datetime.fromisoformat("2024-04-01T20:51:18.077262").replace(tzinfo=timezone.utc)
    return CheckpointResult(
        run_id=RunIdentifier(run_time=utc_datetime),
        run_results={
            ValidationResultIdentifier(
                expectation_suite_identifier=ExpectationSuiteIdentifier(
                    name=SUITE_A,
                ),
                run_id=RunIdentifier(run_name="prod_20240401"),
                batch_identifier=BATCH_ID_A,
            ): ExpectationSuiteValidationResult(
                success=True,
                statistics={"successful_expectations": 3, "evaluated_expectations": 3},
                results=[],
                suite_name=SUITE_A,
            ),
            ValidationResultIdentifier(
                expectation_suite_identifier=ExpectationSuiteIdentifier(
                    name=SUITE_B,
                ),
                run_id=RunIdentifier(run_name="prod_20240402"),
                batch_identifier=BATCH_ID_B,
            ): ExpectationSuiteValidationResult(
                success=True,
                statistics={"successful_expectations": 2, "evaluated_expectations": 2},
                results=[],
                suite_name=SUITE_B,
            ),
        },
        checkpoint_config=Checkpoint(
            name="test-checkpoint",
            validation_definitions=[
                mocker.MagicMock(spec=ValidationDefinition),
                mocker.MagicMock(spec=ValidationDefinition),
            ],
        ),
    )


@pytest.fixture
def checkpoint_result_with_assets(mocker: MockerFixture):
    utc_datetime = datetime.fromisoformat("2024-04-01T20:51:18.077262").replace(tzinfo=timezone.utc)
    return CheckpointResult(
        run_id=RunIdentifier(run_time=utc_datetime),
        run_results={
            ValidationResultIdentifier(
                expectation_suite_identifier=ExpectationSuiteIdentifier(
                    name=SUITE_A,
                ),
                run_id=RunIdentifier(run_name="prod_20240401"),
                batch_identifier=BATCH_ID_A,
            ): ExpectationSuiteValidationResult(
                success=True,
                statistics={"successful_expectations": 3, "evaluated_expectations": 3},
                results=[],
                suite_name=SUITE_A,
                meta={
                    "active_batch_definition": LegacyBatchDefinition(
                        datasource_name="test_environment",
                        data_connector_name="general_azure_data_connector",
                        data_asset_name="asset_1",
                        batch_identifiers=IDDict(
                            {"name": "alex", "timestamp": "20200809", "price": "1000"}
                        ),
                    )
                },
                result_url="www.testing",
            ),
            ValidationResultIdentifier(
                expectation_suite_identifier=ExpectationSuiteIdentifier(
                    name=SUITE_B,
                ),
                run_id=RunIdentifier(run_name="prod_20240402"),
                batch_identifier=BATCH_ID_B,
            ): ExpectationSuiteValidationResult(
                success=True,
                statistics={"successful_expectations": 2, "evaluated_expectations": 2},
                results=[],
                suite_name=SUITE_B,
                meta={
                    "active_batch_definition": LegacyBatchDefinition(
                        datasource_name="test_environment",
                        data_connector_name="general_azure_data_connector",
                        data_asset_name="asset_2_two_wow_whoa_vroom",
                        batch_identifiers=IDDict(
                            {"name": "alex", "timestamp": "20200809", "price": "1000"}
                        ),
                    ),
                },
            ),
        },
        checkpoint_config=mocker.Mock(spec=Checkpoint, name="my_checkpoint"),
    )


@contextmanager
def mock_not_imported_module(
    parent_module: ModuleType, target_name: str, mocker: MockerFixture
) -> Iterator[Never]:
    original = getattr(parent_module, target_name)
    try:
        setattr(parent_module, target_name, mocker.Mock())
        yield getattr(parent_module, target_name)
    finally:
        setattr(parent_module, target_name, original)


class TestAPINotificationAction:
    @pytest.mark.unit
    def test_create_payload(self, mock_context):
        mock_validation_results = []
        expected_payload = {
            "test_suite_name": "my_suite",
            "data_asset_name": "my_schema.my_table",
            "validation_results": [],
        }
        api_notification_action = APINotificationAction(
            name="my_api_notification", url="http://www.example.com"
        )
        payload = api_notification_action.create_payload(
            "my_schema.my_table", "my_suite", mock_validation_results
        )
        assert payload == expected_payload

    @pytest.mark.unit
    def test_run(self, checkpoint_result: CheckpointResult, mocked_posthog):
        url = "http://www.example.com"
        action = APINotificationAction(name="my_action", url=url)

        with mock.patch.object(requests, "post") as mock_post:
            action.run(checkpoint_result=checkpoint_result)

        mock_post.assert_called_once_with(
            url,
            headers={"Content-Type": "application/json"},
            data=[
                {
                    "data_asset_name": BATCH_ID_A,
                    "test_suite_name": SUITE_A,
                    "validation_results": [],
                },
                {
                    "data_asset_name": BATCH_ID_B,
                    "test_suite_name": SUITE_B,
                    "validation_results": [],
                },
            ],
        )


class TestEmailAction:
    @pytest.mark.unit
    def test_equality(self):
        """I know, this one seems silly. But this was a bug."""
        a = EmailAction(
            name="my_action",
            smtp_address="test",
            smtp_port="587",
            receiver_emails="test@gmail.com",
        )
        b = EmailAction(
            name="my_action",
            smtp_address="test",
            smtp_port="587",
            receiver_emails="test@gmail.com",
        )

        assert a == b

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "emails, expected_email_list",
        [
            pytest.param("test1@gmail.com", ["test1@gmail.com"], id="single_email"),
            pytest.param(
                "test1@gmail.com, test2@hotmail.com",
                ["test1@gmail.com", "test2@hotmail.com"],
                id="multiple_emails",
            ),
            pytest.param(
                "test1@gmail.com,test2@hotmail.com",
                ["test1@gmail.com", "test2@hotmail.com"],
                id="multiple_emails_no_space",
            ),
        ],
    )
    def test_run(
        self,
        checkpoint_result: CheckpointResult,
        emails: str,
        expected_email_list: list[str],
        mocked_posthog,
    ):
        action = EmailAction(
            name="my_action",
            smtp_address="test",
            smtp_port="587",
            receiver_emails=emails,
        )

        with mock.patch.object(smtplib, "SMTP") as mock_server:
            out = action.run(checkpoint_result=checkpoint_result)

        mock_send_email = mock_server().sendmail

        # Should contain success/failure in title
        assert (
            f"Subject: {checkpoint_result.checkpoint_config.name}: True"
            in mock_send_email.call_args.args[-1]
        )

        mock_send_email.assert_called_once_with(
            None,
            expected_email_list,
            mock.ANY,
        )
        assert out == {"email_result": "success"}

    @pytest.mark.unit
    def test_run_smptp_address_substitution(
        self, checkpoint_result: CheckpointResult, mocked_posthog
    ):
        config_provider = project_manager.get_config_provider()
        assert isinstance(config_provider, mock.Mock)  # noqa: TID251 # just using for the instance compare

        SMPT_ADDRESS_KEY = "${smtp_address}"
        SMPT_PORT_KEY = "${smtp_port}"
        SENDER_LOGIN_KEY = "${sender_login}"
        SENDER_ALIAS_KEY = "${sender_alias_login}"
        SENDER_PASSWORD_KEY = "${sender_password_login}"
        RECEIVER_EMAILS_KEY = "${receiver_emails}"

        action = EmailAction(
            name="my_action",
            smtp_address=SMPT_ADDRESS_KEY,
            smtp_port=SMPT_PORT_KEY,
            sender_login=SENDER_LOGIN_KEY,
            sender_alias=SENDER_ALIAS_KEY,
            sender_password=SENDER_PASSWORD_KEY,
            receiver_emails=RECEIVER_EMAILS_KEY,
        )

        config_from_uncommitted_config = {
            SMPT_ADDRESS_KEY: "something.com",
            SMPT_PORT_KEY: "123",
            SENDER_LOGIN_KEY: "sender@greatexpectations.io",
            SENDER_ALIAS_KEY: "alias@greatexpectations.io",
            SENDER_PASSWORD_KEY: "sender_password_login",
            RECEIVER_EMAILS_KEY: "foo@greatexpectations.io, bar@great_expectations.io",
        }

        config_provider.substitute_config.side_effect = lambda key: config_from_uncommitted_config[
            key
        ]
        with mock.patch.object(smtplib, "SMTP") as mock_server:
            action.run(checkpoint_result=checkpoint_result)

        mock_server().sendmail.assert_called_once_with(
            config_from_uncommitted_config[SENDER_ALIAS_KEY],
            [
                email.strip()
                for email in config_from_uncommitted_config[RECEIVER_EMAILS_KEY].split(",")
            ],
            mock.ANY,
        )


class TestMicrosoftTeamsNotificationAction:
    @pytest.mark.unit
    def test_run(self, checkpoint_result: CheckpointResult, mocked_posthog):
        action = MicrosoftTeamsNotificationAction(name="my_action", teams_webhook="test")

        with mock.patch.object(Session, "post") as mock_post:
            action.run(checkpoint_result=checkpoint_result)

        mock_post.assert_called_once()

        body = mock_post.call_args.kwargs["json"]["attachments"][0]["content"]["body"]

        assert len(body) == 5

        # Assert header
        assert "Success" in body[0]["columns"][1]["items"][0]["text"]

        # Assert first validation
        assert body[1]["text"] == "Validation Result (1 of 2) ✅"
        assert body[2]["facts"] == [
            {"title": "Data Asset name: ", "value": "--"},
            {"title": "Suite name: ", "value": SUITE_A},
            {
                "title": "Run name: ",
                "value": "prod_20240401",
            },
            {
                "title": "Summary:",
                "value": "*3* of *3* Expectations were met",
            },
        ]

        # Assert second validation
        assert body[3]["text"] == "Validation Result (2 of 2) ✅"
        assert body[4]["facts"] == [
            {"title": "Data Asset name: ", "value": "--"},
            {"title": "Suite name: ", "value": SUITE_B},
            {
                "title": "Run name: ",
                "value": "prod_20240402",
            },
            {
                "title": "Summary:",
                "value": "*2* of *2* Expectations were met",
            },
        ]

    @pytest.mark.unit
    def test_run_webhook_substitution(self, checkpoint_result: CheckpointResult, mocked_posthog):
        config_provider = project_manager.get_config_provider()
        assert isinstance(config_provider, mock.Mock)  # noqa: TID251 # just using for the instance compare

        MS_TEAMS_WEBHOOK_VAR = "${ms_teams_webhook}"
        MS_TEAMS_WEBHOOK_VALUE = "https://my_org.webhook.office.com/webhookb2/abc"

        action = MicrosoftTeamsNotificationAction(
            name="my_action",
            teams_webhook=MS_TEAMS_WEBHOOK_VAR,
        )

        config_from_uncommitted_config = {MS_TEAMS_WEBHOOK_VAR: MS_TEAMS_WEBHOOK_VALUE}

        config_provider.substitute_config.side_effect = lambda key: config_from_uncommitted_config[
            key
        ]
        with mock.patch.object(Session, "post") as mock_send_notification:
            action.run(checkpoint_result=checkpoint_result)

        mock_send_notification.assert_called_once_with(url=MS_TEAMS_WEBHOOK_VALUE, json=mock.ANY)

    @pytest.mark.integration
    def test_run_integration_success(
        self,
        checkpoint_result: CheckpointResult,
    ):
        # Necessary to retrieve config provider
        gx.get_context(mode="ephemeral")

        action = MicrosoftTeamsNotificationAction(
            name="test-action",
            teams_webhook="${GX_MS_TEAMS_WEBHOOK}",  # Set as a secret in GH Actions
        )
        result = action.run(checkpoint_result=checkpoint_result)
        assert result == {
            "microsoft_teams_notification_result": "Microsoft Teams notification succeeded."
        }

    @pytest.mark.integration
    def test_run_integration_failure(
        self,
        checkpoint_result: CheckpointResult,
        caplog,
    ):
        # Necessary to retrieve config provider
        gx.get_context(mode="ephemeral")

        action = MicrosoftTeamsNotificationAction(
            name="test-action",
            teams_webhook="https://fake.office.com/fake",
        )
        with caplog.at_level(logging.WARNING):
            result = action.run(checkpoint_result=checkpoint_result)

        assert result == {"microsoft_teams_notification_result": None}
        assert caplog.records[-1].message.startswith("Failed to connect to Microsoft Teams webhook")


class TestOpsgenieAlertAction:
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "success, message",
        [
            pytest.param(True, "succeeded!", id="success"),
            pytest.param(False, "failed!", id="failure"),
        ],
    )
    def test_run(self, checkpoint_result: CheckpointResult, success: bool, message: str):
        action = OpsgenieAlertAction(name="my_action", api_key="test", notify_on="all")
        checkpoint_result.success = success

        with mock.patch.object(Session, "post") as mock_post:
            output = action.run(checkpoint_result=checkpoint_result)

        mock_post.assert_called_once()
        assert message in mock_post.call_args.kwargs["json"]["message"]
        assert output == {"opsgenie_alert_result": True}


class TestPagerdutyAlertAction:
    @pytest.mark.unit
    def test_run_emits_events(self, checkpoint_result: CheckpointResult, mocker: MockerFixture):
        from great_expectations.checkpoint import actions

        with mock_not_imported_module(actions, "pypd", mocker):
            mock_pypd_event = actions.pypd.EventV2.create
            action = PagerdutyAlertAction(
                name="my_action", api_key="test", routing_key="test", notify_on="all"
            )
            checkpoint_name = checkpoint_result.checkpoint_config.name

            checkpoint_result.success = True
            assert action.run(checkpoint_result=checkpoint_result) == {
                "pagerduty_alert_result": "success"
            }

            checkpoint_result.success = False
            assert action.run(checkpoint_result=checkpoint_result) == {
                "pagerduty_alert_result": "success"
            }

            assert mock_pypd_event.call_count == 2
            mock_pypd_event.assert_has_calls(
                [
                    mock.call(
                        data={
                            "dedup_key": checkpoint_name,
                            "event_action": "trigger",
                            "payload": {
                                "severity": "critical",
                                "source": "Great Expectations",
                                "summary": f"Great Expectations Checkpoint {checkpoint_name} has succeeded",  # noqa: E501 # FIXME CoP
                            },
                            "routing_key": "test",
                        }
                    ),
                    mock.call(
                        data={
                            "dedup_key": checkpoint_name,
                            "event_action": "trigger",
                            "payload": {
                                "severity": "critical",
                                "source": "Great Expectations",
                                "summary": f"Great Expectations Checkpoint {checkpoint_name} has failed",  # noqa: E501 # FIXME CoP
                            },
                            "routing_key": "test",
                        }
                    ),
                ]
            )

    @pytest.mark.skipif(
        not is_library_loadable(library_name="pypd"),
        reason="pypd is not installed",
    )
    @mock.patch("pypd.EventV2.create")
    @pytest.mark.unit
    def test_run_does_not_emit_events(self, mock_pypd_event, checkpoint_result: CheckpointResult):
        action = PagerdutyAlertAction(
            name="my_action", api_key="test", routing_key="test", notify_on="failure"
        )

        checkpoint_result.success = True
        assert action.run(checkpoint_result=checkpoint_result) == {
            "pagerduty_alert_result": "none sent"
        }

        mock_pypd_event.assert_not_called()


class TestSlackNotificationAction:
    @pytest.mark.unit
    def test_equality(self):
        """I kow, this one seems silly. But this was a bug."""
        a = SlackNotificationAction(name="my_action", slack_webhook="test", notify_on="all")
        b = SlackNotificationAction(name="my_action", slack_webhook="test", notify_on="all")

        assert a == b

    @pytest.mark.unit
    def test_run(self, checkpoint_result: CheckpointResult, mocked_posthog):
        action = SlackNotificationAction(name="my_action", slack_webhook="test", notify_on="all")

        with mock.patch.object(Session, "post") as mock_post:
            output = action.run(checkpoint_result=checkpoint_result)

        mock_post.assert_called_once_with(
            url="test",
            headers=None,
            json={
                "blocks": [
                    {"text": {"text": mock.ANY, "type": "plain_text"}, "type": "header"},
                    {
                        "type": "section",
                        "text": {"type": "plain_text", "text": "Runtime: 2024/04/01 08:51 PM"},
                    },
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": (
                                "\n*Asset*: `__no_data_asset_name__`  "
                                f"\n*Expectation Suite*: `{SUITE_A}`"
                                "\n*Summary*: *3* of *3* Expectations were met"
                            ),
                        },
                    },
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": (
                                "\n*Asset*: `__no_data_asset_name__`  "
                                f"\n*Expectation Suite*: `{SUITE_B}`"
                                "\n*Summary*: *2* of *2* Expectations were met"
                            ),
                        },
                    },
                    {"type": "divider"},
                ],
            },
        )

        assert output == {"slack_notification_result": "Slack notification succeeded."}

    @pytest.mark.unit
    def test_run_with_assets(self, checkpoint_result_with_assets: CheckpointResult, mocked_posthog):
        action = SlackNotificationAction(name="my_action", slack_webhook="test", notify_on="all")

        with mock.patch.object(Session, "post") as mock_post:
            output = action.run(checkpoint_result=checkpoint_result_with_assets)

        mock_post.assert_called_once_with(
            url="test",
            headers=None,
            json={
                "blocks": [
                    {"text": {"text": mock.ANY, "type": "plain_text"}, "type": "header"},
                    {
                        "type": "section",
                        "text": {"type": "plain_text", "text": "Runtime: 2024/04/01 08:51 PM"},
                    },
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": (
                                f"\n*Asset*: `asset_1`  \n*Expectation Suite*: {SUITE_A}  "
                                "<www.testing?slack=true|View Results>"
                                "\n*Summary*: *3* of *3* Expectations were met"
                            ),
                        },
                    },
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": (
                                "\n*Asset*: `asset_2_two_wow_whoa_vroom`  "
                                f"\n*Expectation Suite*: `{SUITE_B}`"
                                "\n*Summary*: *2* of *2* Expectations were met"
                            ),
                        },
                    },
                    {"type": "divider"},
                ],
            },
        )

        assert output == {"slack_notification_result": "Slack notification succeeded."}

    @pytest.mark.unit
    def test_grabs_data_docs_pages(
        self, checkpoint_result_with_assets: CheckpointResult, mocked_posthog
    ):
        action = SlackNotificationAction(name="my_action", slack_webhook="test", notify_on="all")

        site_path = "file:///var/folders/vm/wkw13lnd5vsdh3hjmcv9tym00000gn/T/tmpctw4x7yu/validations/my_suite/__none__/20240910T175850.906745Z/foo-bar.html"
        action_context = ActionContext()
        action_context.update(
            action=UpdateDataDocsAction(name="docs_action"),
            action_result={
                ValidationResultIdentifier(
                    expectation_suite_identifier=ExpectationSuiteIdentifier(name="my_suite"),
                    run_id=RunIdentifier(run_name="prod_20240401"),
                    batch_identifier="my_datasource-my_first_asset",
                ): {
                    "local_site": site_path,
                }
            },
        )
        with mock.patch.object(Session, "post") as mock_post:
            output = action.run(
                checkpoint_result=checkpoint_result_with_assets, action_context=action_context
            )

        mock_post.assert_called_once_with(
            url="test",
            headers=None,
            json={
                "blocks": [
                    {"text": {"text": mock.ANY, "type": "plain_text"}, "type": "header"},
                    {
                        "type": "section",
                        "text": {"type": "plain_text", "text": "Runtime: 2024/04/01 08:51 PM"},
                    },
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": (
                                f"\n*Asset*: `asset_1`  \n*Expectation Suite*: {SUITE_A}  "
                                "<www.testing?slack=true|View Results>"
                                "\n*Summary*: *3* of *3* Expectations were met"
                            ),
                        },
                    },
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": mock.ANY,
                        },
                    },
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": (
                                "\n*Asset*: `asset_2_two_wow_whoa_vroom`  "
                                f"\n*Expectation Suite*: `{SUITE_B}`"
                                "\n*Summary*: *2* of *2* Expectations were met"
                            ),
                        },
                    },
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": mock.ANY,
                        },
                    },
                    {"type": "divider"},
                ],
            },
        )

        docs_block_1 = mock_post.call_args.kwargs["json"]["blocks"][3]["text"]["text"]
        docs_block_2 = mock_post.call_args.kwargs["json"]["blocks"][5]["text"]["text"]

        assert "*DataDocs*" in docs_block_1
        assert site_path in docs_block_1
        assert "*DataDocs*" in docs_block_2
        assert site_path in docs_block_2
        assert output == {"slack_notification_result": "Slack notification succeeded."}

    @pytest.mark.unit
    def test_variable_substitution_webhook(self, mock_context, checkpoint_result, mocked_posthog):
        action = SlackNotificationAction(name="my_action", slack_webhook="${SLACK_WEBHOOK}")

        with mock.patch.object(Session, "post"):
            action.run(checkpoint_result)

        mock_context.config_provider.substitute_config.assert_called_once_with("${SLACK_WEBHOOK}")

    @pytest.mark.unit
    def test_variable_substitution_token_and_channel(
        self, mock_context, checkpoint_result, mocked_posthog
    ):
        action = SlackNotificationAction(
            name="my_action", slack_token="${SLACK_TOKEN}", slack_channel="${SLACK_CHANNEL}"
        )

        with mock.patch.object(Session, "post"):
            action.run(checkpoint_result)

        assert mock_context.config_provider.substitute_config.call_count == 2
        mock_context.config_provider.substitute_config.assert_any_call("${SLACK_CHANNEL}")
        mock_context.config_provider.substitute_config.assert_any_call("${SLACK_TOKEN}")


class TestSNSNotificationAction:
    @pytest.mark.unit
    def test_run(self, sns, checkpoint_result: CheckpointResult, mocked_posthog):
        subj_topic = "test-subj"
        created_subj = sns.create_topic(Name=subj_topic)
        arn = created_subj.get("TopicArn")
        action = SNSNotificationAction(
            name="my_action",
            sns_topic_arn=arn,
            sns_message_subject="Subject",
        )

        result = action.run(checkpoint_result=checkpoint_result)
        assert "Successfully posted results" in result["result"]


class TestUpdateDataDocsAction:
    @pytest.mark.unit
    def test_equality(self):
        """I kow, this one seems silly. But this was a bug for other actions."""
        a = UpdateDataDocsAction(name="my_action")
        b = UpdateDataDocsAction(name="my_action")

        assert a == b

    @pytest.mark.unit
    def test_run(self, mocker: MockerFixture, checkpoint_result: CheckpointResult, mocked_posthog):
        # Arrange
        context = mocker.Mock(spec=AbstractDataContext)
        set_context(context)

        site_names = ["site_a", "site_b"]
        site_urls = [
            f"/gx/uncommitted/data_docs/{site_names[0]}/index.html",
            f"/gx/uncommitted/data_docs/{site_names[1]}/index.html",
        ]
        context.get_docs_sites_urls.return_value = [
            {
                "site_url": site_urls[0],
                "site_name": site_names[0],
            },
            {
                "site_url": site_urls[1],
                "site_name": site_names[1],
            },
        ]

        # Act
        action = UpdateDataDocsAction(name="my_action", site_names=site_names)
        res = action.run(checkpoint_result=checkpoint_result)

        # Assert
        validation_identifier_a, validation_identifier_b = tuple(
            checkpoint_result.run_results.keys()
        )
        assert context.build_data_docs.call_count == 2, (
            "Data Docs should be incrementally built (once per validation result)"
        )
        context.build_data_docs.assert_has_calls(
            [
                mock.call(
                    build_index=True,
                    dry_run=False,
                    resource_identifiers=[
                        validation_identifier_a,
                        ExpectationSuiteIdentifier(name=SUITE_A),
                    ],
                    site_names=site_names,
                ),
                mock.call(
                    build_index=True,
                    dry_run=False,
                    resource_identifiers=[
                        validation_identifier_b,
                        ExpectationSuiteIdentifier(name=SUITE_B),
                    ],
                    site_names=site_names,
                ),
            ]
        )
        assert res == {
            validation_identifier_a: {
                site_names[0]: site_urls[0],
                site_names[1]: site_urls[1],
            },
            validation_identifier_b: {
                site_names[0]: site_urls[0],
                site_names[1]: site_urls[1],
            },
        }

    @pytest.mark.cloud
    def test_run_with_cloud(
        self, mocker: MockerFixture, checkpoint_result: CheckpointResult, mocked_posthog
    ):
        # Arrange
        context = mocker.Mock(spec=CloudDataContext)
        set_context(context)

        site_names = ["site_a", "site_b"]
        site_urls = [
            f"http://app.greatexpectations.io/data_docs/{site_names[0]}",
            f"http://app.greatexpectations.io/data_docs/{site_names[1]}",
        ]
        context.get_docs_sites_urls.return_value = [
            {
                "site_url": site_urls[0],
                "site_name": site_names[0],
            },
            {
                "site_url": site_urls[1],
                "site_name": site_names[1],
            },
        ]

        # Act
        action = UpdateDataDocsAction(name="my_docs_action", site_names=site_names)
        res = action.run(checkpoint_result=checkpoint_result)

        # Assert
        validation_identifier_a, validation_identifier_b = tuple(
            checkpoint_result.run_results.keys()
        )
        assert context.build_data_docs.call_count == 2, (
            "Data Docs should be incrementally built (once per validation result)"
        )
        context.build_data_docs.assert_has_calls(
            [
                mock.call(
                    build_index=True,
                    dry_run=False,
                    resource_identifiers=[
                        validation_identifier_a,
                        GXCloudIdentifier(
                            resource_type=GXCloudRESTResource.EXPECTATION_SUITE,
                            resource_name=SUITE_A,
                        ),
                    ],
                    site_names=site_names,
                ),
                mock.call(
                    build_index=True,
                    dry_run=False,
                    resource_identifiers=[
                        validation_identifier_b,
                        GXCloudIdentifier(
                            resource_type=GXCloudRESTResource.EXPECTATION_SUITE,
                            resource_name=SUITE_B,
                        ),
                    ],
                    site_names=site_names,
                ),
            ]
        )
        assert res == {
            validation_identifier_a: {},
            validation_identifier_b: {},
        }


class TestCustomActions:
    @pytest.mark.unit
    def test_custom_action_shadows_existing_type(self):
        with pytest.raises(ValidationActionAlreadyRegisteredError):

            class CustomSlackAction(ValidationAction):
                type: Literal["slack"] = "slack"  # Shadows existing value



from dataclasses import dataclass
from typing import List, Optional, Tuple

from tests.integration.backend_dependencies import BackendDependencies


@dataclass
class IntegrationTestFixture:
    """IntegrationTestFixture

    Configurations for integration tests are defined as IntegrationTestFixture dataclass objects.

    Individual tests can also be run by setting the '-k' flag and referencing the name of test, like the following example:
    pytest -v --docs-tests -k "test_docs[migration_guide_spark_v2_api]" tests/integration/test_script_runner.py

    Args:
        name: Name for integration test. Individual tests can be run by using the -k option and specifying the name of the test.
        user_flow_script: Required script for integration test.
        backend_dependencies: Flag allows you to tie an individual test with a BackendDependency. Allows for tests to be run / disabled using cli flags (like --aws which enables AWS integration tests). If no backends are required, must explicitly set to empty list.
        data_context_dir: Path of great_expectations/ that is used in the test.
        data_dir: Folder that contains data used in the test.
        other_files: other files (like credential information) to copy into the test environment. These are presented as Tuple(path_to_source_file, path_to_target_file), where path_to_target_file is relative to the test_script.py file in our test environment
        util_script: Path of optional util script that is used in test script (for loading test_specific methods like load_data_into_test_database())
    """  # noqa: E501 # FIXME CoP

    name: str
    user_flow_script: str
    backend_dependencies: List[BackendDependencies]
    data_context_dir: Optional[str] = None
    data_dir: Optional[str] = None
    other_files: Optional[Tuple[Tuple[str, str]]] = None
    util_script: Optional[str] = None



from __future__ import annotations

from typing import Literal, Union

from great_expectations._docs_decorators import public_api
from great_expectations.compatibility.pydantic import PostgresDsn
from great_expectations.datasource.fluent.config_str import ConfigStr
from great_expectations.datasource.fluent.sql_datasource import SQLDatasource


@public_api
class AuroraDatasource(SQLDatasource):
    """Adds an aurora datasource to the data context.

    Args:
        name: The name of this aurora datasource.
        connection_string: The connection string used to connect to the postgres database.
            For example: "postgresql://<username>:<password>@<cluster-endpoint>.amazonaws.com:<port>/<database_name>"
        assets: An optional dictionary whose keys are TableAsset or QueryAsset names and whose
            values are TableAsset or QueryAsset objects.
    """

    type: Literal["aurora"] = "aurora"  # type: ignore[assignment]
    connection_string: Union[ConfigStr, PostgresDsn]



from __future__ import annotations

import uuid
from pathlib import Path
from typing import TYPE_CHECKING, Generator

import pytest

from great_expectations import ValidationDefinition
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.datasource.fluent.data_connector.filesystem_data_connector import (
    normalize_directory_path,
)

if TYPE_CHECKING:
    import pandas as pd

    from great_expectations.checkpoint.checkpoint import Checkpoint, CheckpointResult
    from great_expectations.core import ExpectationSuite
    from great_expectations.data_context import CloudDataContext
    from great_expectations.datasource.fluent import (
        SparkFilesystemDatasource,
    )
    from great_expectations.datasource.fluent.data_asset.path.spark.csv_asset import CSVAsset


@pytest.fixture(scope="module")
def csv_path() -> Path:
    return Path("data.csv")


@pytest.fixture(scope="module")
def base_dir(tmp_path: Path, csv_path: Path, test_data: pd.DataFrame) -> Path:
    dir_path = tmp_path / "data"
    dir_path.mkdir()
    csv_path = dir_path / csv_path
    test_data.to_csv(csv_path)
    return dir_path


@pytest.fixture(scope="module")
def updated_base_dir(tmp_path: Path) -> Path:
    dir_path = tmp_path / "other_data"
    dir_path.mkdir()
    return dir_path


@pytest.fixture(scope="module")
def datasource(
    context: CloudDataContext,
    datasource_name: str,
    base_dir: Path,
    updated_base_dir: Path,
) -> SparkFilesystemDatasource:
    """Test Adding and Updating the Datasource associated with this module.
    Note: There is no need to test Get or Delete Datasource.
    Those assertions can be found in the datasource_name fixture."""
    original_base_dir = base_dir

    datasource = context.data_sources.add_spark_filesystem(
        name=datasource_name, base_directory=original_base_dir
    )

    datasource.base_directory = normalize_directory_path(updated_base_dir, context.root_directory)
    datasource = context.data_sources.add_or_update_spark_filesystem(datasource=datasource)
    assert datasource.base_directory == updated_base_dir, (
        "The datasource was not updated in the previous method call."
    )

    datasource.base_directory = normalize_directory_path(original_base_dir, context.root_directory)
    datasource = context.data_sources.add_or_update_spark_filesystem(datasource=datasource)
    assert datasource.base_directory == original_base_dir, (
        "The datasource was not updated in the previous method call."
    )
    return datasource


@pytest.fixture(scope="module")
def csv_asset(
    datasource: SparkFilesystemDatasource,
) -> Generator[CSVAsset, None, None]:
    """Test CSVAsset lifecycle"""
    asset_name = f"asset_{uuid.uuid4().hex}"
    yield datasource.add_csv_asset(name=asset_name, header=True, infer_schema=True)
    datasource.delete_asset(name=asset_name)
    with pytest.raises(LookupError):
        datasource.get_asset(name=asset_name)


@pytest.fixture(scope="module")
def batch_definition(csv_asset: CSVAsset, csv_path: Path) -> BatchDefinition:
    batch_def_name = f"batch_def_{uuid.uuid4().hex}"
    return csv_asset.add_batch_definition_path(
        name=batch_def_name,
        path=csv_path,
    )


@pytest.fixture(scope="module")
def validation_definition(
    context: CloudDataContext,
    expectation_suite: ExpectationSuite,
    batch_definition: BatchDefinition,
) -> Generator[ValidationDefinition, None, None]:
    validation_def_name = f"val_def_{uuid.uuid4().hex}"
    yield context.validation_definitions.add(
        ValidationDefinition(
            name=validation_def_name,
            data=batch_definition,
            suite=expectation_suite,
        )
    )
    context.validation_definitions.delete(name=validation_def_name)


@pytest.fixture(scope="module")
def checkpoint(
    checkpoint: Checkpoint,
    validation_definition: ValidationDefinition,
) -> Checkpoint:
    checkpoint.validation_definitions = [validation_definition]
    checkpoint.save()
    return checkpoint


@pytest.mark.cloud
def test_checkpoint_run(checkpoint: Checkpoint):
    """Test running a Checkpoint that was created using the entities defined in this module."""
    checkpoint_result: CheckpointResult = checkpoint.run()
    assert checkpoint_result.success



from __future__ import annotations

from typing import Literal, Optional, Union

from great_expectations.compatibility.pydantic import Field
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent.data_asset.path.directory_asset import (
    DirectoryDataAsset,
)
from great_expectations.datasource.fluent.data_asset.path.file_asset import FileDataAsset
from great_expectations.datasource.fluent.data_asset.path.spark.spark_generic import (
    _SparkGenericFilePathAssetMixin,
)
from great_expectations.datasource.fluent.serializable_types.pyspark import (
    SerializableStructType,  # noqa: TC001  # pydantic uses type at runtime
)


class JSONAssetBase(_SparkGenericFilePathAssetMixin):
    # vvv spark parameters for pyspark.sql.DataFrameReader.json() (ordered as in pyspark v3.4.0) appear in comment above  # noqa: E501 # FIXME CoP
    # parameter for reference (from https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L309)
    # path: Union[str, List[str], RDD[str]],
    # NA - path determined by asset
    # schema: Optional[Union[StructType, str]] = None,
    # schema shadows pydantic BaseModel attribute
    spark_schema: Optional[Union[SerializableStructType, str]] = Field(None, alias="schema")
    # primitivesAsString: Optional[Union[bool, str]] = None,
    primitives_as_string: Optional[Union[bool, str]] = Field(None, alias="primitivesAsString")
    # prefersDecimal: Optional[Union[bool, str]] = None,
    prefers_decimal: Optional[Union[bool, str]] = Field(None, alias="prefersDecimal")
    # allowComments: Optional[Union[bool, str]] = None,
    allow_comments: Optional[Union[bool, str]] = Field(None, alias="allowComments")
    # allowUnquotedFieldNames: Optional[Union[bool, str]] = None,
    allow_unquoted_field_names: Optional[Union[bool, str]] = Field(
        None, alias="allowUnquotedFieldNames"
    )
    # allowSingleQuotes: Optional[Union[bool, str]] = None,
    allow_single_quotes: Optional[Union[bool, str]] = Field(None, alias="allowSingleQuotes")
    # allowNumericLeadingZero: Optional[Union[bool, str]] = None,
    allow_numeric_leading_zero: Optional[Union[bool, str]] = Field(
        None, alias="allowNumericLeadingZero"
    )
    # allowBackslashEscapingAnyCharacter: Optional[Union[bool, str]] = None,
    allow_backslash_escaping_any_character: Optional[Union[bool, str]] = Field(
        None, alias="allowBackslashEscapingAnyCharacter"
    )
    # mode: Optional[str] = None,
    mode: Optional[Literal["PERMISSIVE", "DROPMALFORMED", "FAILFAST"]] = None
    # columnNameOfCorruptRecord: Optional[str] = None,
    column_name_of_corrupt_record: Optional[str] = Field(None, alias="columnNameOfCorruptRecord")
    # dateFormat: Optional[str] = None,
    date_format: Optional[str] = Field(None, alias="dateFormat")
    # timestampFormat: Optional[str] = None,
    timestamp_format: Optional[str] = Field(None, alias="timestampFormat")
    # multiLine: Optional[Union[bool, str]] = None,
    multi_line: Optional[Union[bool, str]] = Field(None, alias="multiLine")
    # allowUnquotedControlChars: Optional[Union[bool, str]] = None,
    allow_unquoted_control_chars: Optional[Union[bool, str]] = Field(
        None, alias="allowUnquotedControlChars"
    )
    # lineSep: Optional[str] = None,
    line_sep: Optional[str] = Field(None, alias="lineSep")
    # samplingRatio: Optional[Union[float, str]] = None,
    sampling_ratio: Optional[Union[float, str]] = Field(None, alias="samplingRatio")
    # dropFieldIfAllNull: Optional[Union[bool, str]] = None,
    drop_field_if_all_null: Optional[Union[bool, str]] = Field(None, alias="dropFieldIfAllNull")
    # encoding: Optional[str] = None,
    encoding: Optional[str] = None
    # locale: Optional[str] = None,
    locale: Optional[str] = None
    # pathGlobFilter: Optional[Union[bool, str]] = None,
    # Inherited from _SparkGenericFilePathAssetMixin
    # recursiveFileLookup: Optional[Union[bool, str]] = None,
    # Inherited from _SparkGenericFilePathAssetMixin
    # modifiedBefore: Optional[Union[bool, str]] = None,
    # Inherited from _SparkGenericFilePathAssetMixin
    # modifiedAfter: Optional[Union[bool, str]] = None,
    # Inherited from _SparkGenericFilePathAssetMixin
    # allowNonNumericNumbers: Optional[Union[bool, str]] = None,
    allow_non_numeric_numbers: Optional[Union[bool, str]] = Field(
        None, alias="allowNonNumericNumbers"
    )
    # ^^^ spark parameters for pyspark.sql.DataFrameReader.json() (ordered as in pyspark v3.4.0)

    # vvv Docs <> Source Code mismatch
    # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-json.html
    # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L309
    # timezone: str = Field(alias="timeZone")
    # timestamp_ntz_format: str = Field(
    #     "yyyy-MM-dd'T'HH:mm:ss[.SSS]", alias="timestampNTZFormat"
    # )
    # enable_date_time_parsing_fallback: bool = Field(
    #     alias="enableDateTimeParsingFallback"
    # )
    # ^^^ Docs <> Source Code mismatch

    class Config:
        extra = "forbid"

        allow_population_by_field_name = True

    @classmethod
    @override
    def _get_reader_method(cls) -> str:
        return "json"

    @override
    def _get_reader_options_include(self) -> set[str]:
        return (
            super()
            ._get_reader_options_include()
            .union(
                {
                    "primitives_as_string",
                    "prefers_decimal",
                    "allow_comments",
                    "allow_unquoted_field_names",
                    "allow_single_quotes",
                    "allow_numeric_leading_zero",
                    "allow_backslash_escaping_any_character",
                    "mode",
                    "column_name_of_corrupt_record",
                    "date_format",
                    "timestamp_format",
                    "multi_line",
                    "allow_unquoted_control_chars",
                    "line_sep",
                    "sampling_ratio",
                    "drop_field_if_all_null",
                    "encoding",
                    "locale",
                    "allow_non_numeric_numbers",
                    # Inherited vvv
                    # "pathGlobFilter",
                    # "recursiveFileLookup",
                    # "modifiedBefore",
                    # "modifiedAfter",
                    # Inherited ^^^
                    # vvv Docs <> Source Code mismatch
                    # The following parameters are mentioned in https://spark.apache.org/docs/latest/sql-data-sources-json.html
                    # however do not appear in the source code https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L309
                    # "enableDateTimeParsingFallback",
                    # "timeZone",
                    # "timestampNTZFormat",
                    # ^^^ Docs <> Source Code mismatch
                }
            )
        )


class JSONAsset(FileDataAsset, JSONAssetBase):
    type: Literal["json"] = "json"


class DirectoryJSONAsset(DirectoryDataAsset, JSONAssetBase):
    type: Literal["directory_json"] = "directory_json"

    @classmethod
    @override
    def _get_reader_method(cls) -> str:
        return "json"

    @override
    def _get_reader_options_include(self) -> set[str]:
        """These options are available as of spark v3.4.0

        See https://spark.apache.org/docs/latest/sql-data-sources-json.html for more info.
        """
        return (
            super()._get_reader_options_include()
            | super(DirectoryDataAsset, self)._get_reader_options_include()
        )



from __future__ import annotations

import logging
import os
from typing import Any, Dict, Iterator, List, cast
from unittest import mock

import pytest

import great_expectations.exceptions as ge_exceptions
from great_expectations.compatibility import google
from great_expectations.datasource.fluent import (
    PandasGoogleCloudStorageDatasource,
)
from great_expectations.datasource.fluent.data_asset.path.pandas.generated_assets import CSVAsset
from great_expectations.datasource.fluent.dynamic_pandas import PANDAS_VERSION

logger = logging.getLogger(__file__)


if not google.storage:
    pytest.skip(
        'Could not import "storage" from google.cloud in configured_asset_gcs_data_connector.py',
        allow_module_level=True,
    )


# apply markers to entire test module
pytestmark = [
    pytest.mark.skipif(
        PANDAS_VERSION < 1.2, reason=f"Fluent pandas not supported on {PANDAS_VERSION}"
    )
]


class MockGCSClient:
    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def list_blobs(
        self,
        bucket_or_name,
        max_results=None,
        prefix=None,
        delimiter=None,
        **kwargs,
    ) -> Iterator:
        return iter([])


def _build_pandas_gcs_datasource(
    gcs_options: Dict[str, Any] | None = None,
) -> PandasGoogleCloudStorageDatasource:
    gcs_client: google.Client = cast("google.Client", MockGCSClient())
    pandas_gcs_datasource = PandasGoogleCloudStorageDatasource(
        name="pandas_gcs_datasource",
        bucket_or_name="test_bucket",
        gcs_options=gcs_options or {},
    )
    pandas_gcs_datasource._gcs_client = gcs_client
    return pandas_gcs_datasource


@pytest.fixture
def pandas_gcs_datasource() -> PandasGoogleCloudStorageDatasource:
    pandas_gcs_datasource: PandasGoogleCloudStorageDatasource = _build_pandas_gcs_datasource()
    return pandas_gcs_datasource


@pytest.fixture
def object_keys() -> List[str]:
    return [
        "alex_20200809_1000.csv",
        "eugene_20200809_1500.csv",
        "james_20200811_1009.csv",
        "abe_20200809_1040.csv",
        "will_20200809_1002.csv",
        "james_20200713_1567.csv",
        "eugene_20201129_1900.csv",
        "will_20200810_1001.csv",
        "james_20200810_1003.csv",
        "alex_20200819_1300.csv",
    ]


@pytest.mark.unit
def test_construct_pandas_gcs_datasource_without_gcs_options():
    google_cred_file = os.getenv("GOOGLE_APPLICATION_CREDENTIALS")
    if not google_cred_file:
        pytest.skip('No "GOOGLE_APPLICATION_CREDENTIALS" environment variable found.')

    pandas_gcs_datasource = PandasGoogleCloudStorageDatasource(
        name="pandas_gcs_datasource",
        bucket_or_name="test_bucket",
        gcs_options={},
    )
    gcs_client: google.Client = pandas_gcs_datasource._get_gcs_client()
    assert gcs_client is not None
    assert pandas_gcs_datasource.name == "pandas_gcs_datasource"


@pytest.mark.unit
@mock.patch(
    "great_expectations.datasource.fluent.data_asset.data_connector.google_cloud_storage_data_connector.list_gcs_keys"
)
@mock.patch("google.oauth2.service_account.Credentials.from_service_account_file")
@mock.patch("google.cloud.storage.Client")
def test_construct_pandas_gcs_datasource_with_filename_in_gcs_options(
    mock_gcs_client, mock_gcs_service_account_credentials, mock_list_keys
):
    pandas_gcs_datasource = PandasGoogleCloudStorageDatasource(
        name="pandas_gcs_datasource",
        bucket_or_name="test_bucket",
        gcs_options={
            "filename": "my_filename.csv",
        },
    )
    gcs_client: google.Client = pandas_gcs_datasource._get_gcs_client()
    assert gcs_client is not None
    assert pandas_gcs_datasource.name == "pandas_gcs_datasource"


@pytest.mark.unit
@mock.patch(
    "great_expectations.datasource.fluent.data_asset.data_connector.google_cloud_storage_data_connector.list_gcs_keys"
)
@mock.patch("google.oauth2.service_account.Credentials.from_service_account_info")
@mock.patch("google.cloud.storage.Client")
def test_construct_pandas_gcs_datasource_with_info_in_gcs_options(
    mock_gcs_client, mock_gcs_service_account_credentials, mock_list_keys
):
    pandas_gcs_datasource = PandasGoogleCloudStorageDatasource(
        name="pandas_gcs_datasource",
        bucket_or_name="test_bucket",
        gcs_options={
            "info": "{my_csv: my_content,}",
        },
    )
    gcs_client: google.Client = pandas_gcs_datasource._get_gcs_client()
    assert gcs_client is not None
    assert pandas_gcs_datasource.name == "pandas_gcs_datasource"


@pytest.mark.unit
@mock.patch(
    "great_expectations.datasource.fluent.data_asset.data_connector.google_cloud_storage_data_connector.list_gcs_keys"
)
@mock.patch("google.cloud.storage.Client")
def test_add_csv_asset_to_datasource(
    mock_gcs_client,
    mock_list_keys,
    object_keys: List[str],
    pandas_gcs_datasource: PandasGoogleCloudStorageDatasource,
):
    mock_list_keys.return_value = object_keys
    asset = pandas_gcs_datasource.add_csv_asset(
        name="csv_asset",
    )
    assert asset.name == "csv_asset"


@pytest.mark.unit
@mock.patch(
    "great_expectations.datasource.fluent.data_asset.data_connector.google_cloud_storage_data_connector.list_gcs_keys"
)
@mock.patch("google.cloud.storage.Client")
def test_construct_csv_asset_directly(mock_gcs_client, mock_list_keys, object_keys: List[str]):
    mock_list_keys.return_value = object_keys
    asset = CSVAsset(  # type: ignore[call-arg] # FIXME CoP
        name="csv_asset",
    )
    assert asset.name == "csv_asset"


@pytest.mark.unit
@mock.patch(
    "great_expectations.datasource.fluent.data_asset.data_connector.google_cloud_storage_data_connector.list_gcs_keys"
)
@mock.patch("google.cloud.storage.Client")
def test_csv_asset_with_batching_regex_named_parameters(
    mock_gcs_client,
    mock_list_keys,
    object_keys: List[str],
    pandas_gcs_datasource: PandasGoogleCloudStorageDatasource,
):
    mock_list_keys.return_value = object_keys
    asset = pandas_gcs_datasource.add_csv_asset(
        name="csv_asset",
    )
    batching_regex = r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"

    batch_def = asset.add_batch_definition_monthly(name="batch def", regex=batching_regex)
    options = asset.get_batch_parameters_keys(partitioner=batch_def.partitioner)
    assert options == ("path", "year", "month")


@pytest.mark.unit
@mock.patch(
    "great_expectations.datasource.fluent.data_asset.data_connector.google_cloud_storage_data_connector.list_gcs_keys"
)
@mock.patch("google.cloud.storage.Client")
def test_csv_asset_with_non_string_batching_regex_named_parameters(
    mock_gcs_client,
    mock_list_keys,
    object_keys: List[str],
    pandas_gcs_datasource: PandasGoogleCloudStorageDatasource,
):
    mock_list_keys.return_value = object_keys
    asset = pandas_gcs_datasource.add_csv_asset(
        name="csv_asset",
    )
    with pytest.raises(ge_exceptions.InvalidBatchRequestError):
        # price is an int which will raise an error
        asset.build_batch_request({"name": "alex", "timestamp": "1234567890", "price": 1300})


@pytest.mark.unit
@mock.patch(
    "great_expectations.datasource.fluent.data_asset.data_connector.google_cloud_storage_data_connector.list_gcs_keys"
)
@mock.patch("google.cloud.storage.Client")
def test_add_csv_asset_with_recursive_file_discovery_to_datasource(
    mock_gcs_client,
    mock_list_keys,
    object_keys: List[str],
    pandas_gcs_datasource: PandasGoogleCloudStorageDatasource,
):
    """
    Tests that the gcs_recursive_file_discovery-flag is passed on
    to the list_keys-function as the recursive-parameter

    This makes the list_keys-function search and return files also
    from sub-directories on GCS, not just the files in the folder
    specified with the abs_name_starts_with-parameter
    """
    mock_list_keys.return_value = object_keys
    pandas_gcs_datasource.add_csv_asset(
        name="csv_asset",
        gcs_recursive_file_discovery=True,
    )
    assert "recursive" in mock_list_keys.call_args.kwargs
    assert mock_list_keys.call_args.kwargs["recursive"] is True



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from great_expectations.compatibility.typing_extensions import override

logger = logging.getLogger(__name__)

from great_expectations.render.renderer.renderer import Renderer

if TYPE_CHECKING:
    from great_expectations.checkpoint.checkpoint import CheckpointResult
    from great_expectations.core.expectation_validation_result import (
        ExpectationSuiteValidationResult,
    )


class OpsgenieRenderer(Renderer):
    @override
    def render(self, checkpoint_result: CheckpointResult):
        text_blocks: list[str] = []
        for run_result in checkpoint_result.run_results.values():
            text_block = self._render_validation_result(result=run_result)
            text_blocks.append(text_block)

        return self._concatenate_text_blocks(
            checkpoint_result=checkpoint_result, text_blocks=text_blocks
        )

    def _render_validation_result(self, result: ExpectationSuiteValidationResult) -> str:
        suite_name = result.suite_name
        data_asset_name = result.asset_name or "__no_data_asset_name__"
        n_checks_succeeded = result.statistics["successful_expectations"]
        n_checks = result.statistics["evaluated_expectations"]
        run_id = result.meta.get("run_id", "__no_run_id__")
        batch_id = result.batch_id or "__no_batch_id__"
        check_details_text = f"{n_checks_succeeded} of {n_checks} expectations were met"

        if result.success:
            status = "Success 🎉"
        else:
            status = "Failed ❌"

        return f"""Batch Validation Status: {status}
Expectation Suite Name: {suite_name}
Data Asset Name: {data_asset_name}
Run ID: {run_id}
Batch ID: {batch_id}
Summary: {check_details_text}"""

    def _concatenate_text_blocks(
        self, checkpoint_result: CheckpointResult, text_blocks: list[str]
    ) -> str:
        checkpoint_name = checkpoint_result.checkpoint_config.name
        success = checkpoint_result.success
        run_id = checkpoint_result.run_id.run_time

        title = f"Checkpoint: {checkpoint_name} - Run ID: {run_id}"
        status = "Status: Failed ❌" if not success else "Status: Success 🎉"
        return f"{title}\n{status}\n\n" + "\n\n".join(text_blocks)

    def _custom_blocks(self, evr):
        return None

    def _get_report_element(self, docs_link):
        return None



from typing import Annotated, Any, ClassVar, Final, Generic, TypeVar

from typing_extensions import dataclass_transform, get_args

from great_expectations.compatibility.pydantic import BaseModel, Field, ModelMetaclass, StrictStr
from great_expectations.metrics.metric_results import MetricResult
from great_expectations.validator.metric_configuration import (
    MetricConfiguration,
    MetricConfigurationID,
)

ALLOWABLE_METRIC_MIXINS: Final[int] = 1

NonEmptyString = Annotated[StrictStr, Field(min_length=1)]


class MixinTypeError(TypeError):
    def __init__(self, class_name: str, mixin_superclass_name: str) -> None:
        super().__init__(
            f"`{class_name}` must use a single `{mixin_superclass_name}` subclass mixin."
        )


class MissingAttributeError(AttributeError):
    def __init__(self, class_name: str, attribute_name: str) -> None:
        super().__init__(f"`{class_name}` must define `{attribute_name}` attribute.")


class UnregisteredMetricError(ValueError):
    def __init__(self, metric_name: str) -> None:
        super().__init__(f"Metric `{metric_name}` was not found in the registry.")


class AbstractClassInstantiationError(TypeError):
    def __init__(self, class_name: str) -> None:
        super().__init__(f"Cannot instantiate abstract class `{class_name}`.")


class EmptyStrError(ValueError):
    def __init__(self, param_name) -> None:
        super().__init__("{param_name} must be a non-empty string.")


@dataclass_transform()
class MetaMetric(ModelMetaclass):
    """Metaclass for Metric classes that maintains a registry of all concrete Metric types."""

    _registry: dict[str, type["Metric"]] = {}

    def __new__(cls, name, bases, attrs, **kwargs):
        register_cls = super().__new__(cls, name, bases, attrs)
        # Don't register the base Metric class
        if name != "Metric":
            metric_name = attrs.get("name")
            # Some subclasses of metric may not have a name
            # Those classes will not be registered
            if metric_name:
                MetaMetric._registry[metric_name] = register_cls
        return register_cls

    @classmethod
    def get_registered_metric_class_from_metric_name(cls, metric_name: str) -> type["Metric"]:
        """Returns the registered Metric class for a given metric name."""
        try:
            return cls._registry[metric_name]
        except KeyError:
            raise UnregisteredMetricError(metric_name)


_MetricResult = TypeVar("_MetricResult", bound=MetricResult)


class Metric(Generic[_MetricResult], BaseModel, metaclass=MetaMetric):
    """The abstract base class for defining all metrics.

    A Metric represents a measurable property that can be computed over a specific domain
    of data (e.g., a column, table, or column pair). All concrete metric implementations
    must inherit from this class and specify their domain type as a mixin.

    Examples:
        A metric for a single column max value:

        >>> class ColumnMaxResult(MetricResult[int]): ...
        >>>
        >>> class ColumnMax(Metric[ColumnMaxResult]):
        ...     ...

        A metric for a single batch row count value:

        >>> class BatchRowCountResult(MetricResult[int]): ...
        >>>
        >>> class BatchRowCount(Metric[BatchRowCountResult]):
        ...     ...

    Notes:
        - The Metric class cannot be instantiated directly - it must be subclassed.
        - Once Metrics are instantiated, they are immutable.

    See Also:
        MetricConfiguration: Configuration class for metric computation
    """

    # we wouldn't mind removing this `name` attribute
    # it's currently our only hook into the legacy metrics system
    name: ClassVar[StrictStr]

    # we use this list to build a dictionary of domain-specific fields
    # by introspecting inheriting models for these keys
    _domain_fields: ClassVar[tuple[str, ...]] = (
        "column",
        "column_list",
        "column_A",
        "column_B",
        "condition_parser",
        "ignore_row_if",
        "row_condition",
    )

    class Config:
        arbitrary_types_allowed = True
        frozen = True

    def __new__(cls, *args, **kwargs):
        if cls is Metric:
            raise AbstractClassInstantiationError(cls.__name__)
        return super().__new__(cls)

    def metric_id_for_batch(self, batch_id: str) -> MetricConfigurationID:
        return self.config(batch_id=batch_id).id

    def config(self, batch_id: str) -> MetricConfiguration:
        # This class is frozen so Metric._to_config will always return the same value
        # when the same batch_id is passed in.
        if not batch_id:
            raise EmptyStrError("batch_id")

        config = Metric._to_config(
            instance_class=self.__class__,
            batch_id=batch_id,
            metric_value_set=list(self.dict().items()),
        )
        return config

    @classmethod
    def get_metric_result_type(cls) -> type[_MetricResult]:
        """Returns the MetricResult type for this Metric."""
        return get_args(getattr(cls, "__orig_bases__", [])[0])[0]

    @staticmethod
    def _to_config(
        instance_class: type["Metric"], batch_id: str, metric_value_set: list[tuple[str, Any]]
    ) -> MetricConfiguration:
        """Returns a MetricConfiguration instance for this Metric."""
        metric_domain_kwargs = {}
        metric_value_kwargs = {}
        metric_values = dict(metric_value_set)
        metric_fields = Metric.__fields__
        domain_fields = {
            field_name: field_info
            for field_name, field_info in instance_class.__fields__.items()
            if field_name in Metric._domain_fields and field_name not in metric_fields
        }
        for field_name, field_info in domain_fields.items():
            metric_domain_kwargs[field_name] = metric_values.get(field_name, field_info.default)
        # Set the batch_id with the passed in value
        metric_domain_kwargs["batch_id"] = batch_id

        value_fields = {
            field_name: field_info
            for field_name, field_info in instance_class.__fields__.items()
            if field_name not in Metric._domain_fields and field_name not in metric_fields
        }
        for field_name, field_info in value_fields.items():
            metric_value_kwargs[field_name] = metric_values.get(field_name, field_info.default)

        return MetricConfiguration(
            metric_name=instance_class.name,
            metric_domain_kwargs=metric_domain_kwargs,
            metric_value_kwargs=metric_value_kwargs,
        )



from __future__ import annotations

import logging
import re
from typing import Dict, List, Match, Optional, Tuple

logger = logging.getLogger(__name__)


class RegExParser:
    def __init__(
        self,
        regex_pattern: re.Pattern,
        unnamed_regex_group_prefix: str = "unnamed_group_",
    ) -> None:
        self._num_all_matched_group_values: int = regex_pattern.groups

        # Check for `(?P<name>)` named group syntax
        self._group_name_to_index_dict: Dict[str, int] = dict(regex_pattern.groupindex)

        self._regex_pattern: re.Pattern = regex_pattern
        self._unnamed_regex_group_prefix: str = unnamed_regex_group_prefix

    def get_num_all_matched_group_values(self) -> int:
        return self._num_all_matched_group_values

    def get_named_group_name_to_group_index_mapping(self) -> Dict[str, int]:
        return self._group_name_to_index_dict

    def get_matches(self, target: str) -> Optional[Match[str]]:
        return self._regex_pattern.match(target)

    def get_all_group_names_to_group_indexes_bidirectional_mappings(
        self,
    ) -> Tuple[Dict[str, int], Dict[int, str]]:
        named_group_index_to_group_name_mapping: Dict[int, str] = dict(
            zip(
                self._group_name_to_index_dict.values(),
                self._group_name_to_index_dict.keys(),
            )
        )

        idx: int
        common_group_indexes: List[int] = list(
            filter(
                lambda idx: idx not in self._group_name_to_index_dict.values(),
                range(1, self._num_all_matched_group_values + 1),
            )
        )

        group_idx: int
        common_group_index_to_group_name_mapping: Dict[int, str] = {
            group_idx: f"{self._unnamed_regex_group_prefix}{group_idx}"
            for group_idx in common_group_indexes
        }

        all_group_index_to_group_name_mapping: Dict[int, str] = {
            **named_group_index_to_group_name_mapping,
            **common_group_index_to_group_name_mapping,
        }

        element: Tuple[int, str]
        # noinspection PyTypeChecker
        all_group_index_to_group_name_mapping = dict(
            sorted(
                all_group_index_to_group_name_mapping.items(),
                key=lambda element: element[0],
                reverse=False,
            )
        )

        all_group_name_to_group_index_mapping: Dict[str, int] = dict(
            zip(
                all_group_index_to_group_name_mapping.values(),
                all_group_index_to_group_name_mapping.keys(),
            )
        )

        return (
            all_group_name_to_group_index_mapping,
            all_group_index_to_group_name_mapping,
        )

    def get_all_group_name_to_group_index_mapping(self) -> Dict[str, int]:
        all_group_names_to_group_indexes_bidirectional_mappings: Tuple[
            Dict[str, int], Dict[int, str]
        ] = self.get_all_group_names_to_group_indexes_bidirectional_mappings()
        all_group_name_to_group_index_mapping: Dict[str, int] = (
            all_group_names_to_group_indexes_bidirectional_mappings[0]
        )
        return all_group_name_to_group_index_mapping

    def get_all_group_index_to_group_name_mapping(self) -> Dict[int, str]:
        all_group_names_to_group_indexes_bidirectional_mappings: Tuple[
            Dict[str, int], Dict[int, str]
        ] = self.get_all_group_names_to_group_indexes_bidirectional_mappings()
        all_group_index_to_group_name_mapping: Dict[int, str] = (
            all_group_names_to_group_indexes_bidirectional_mappings[1]
        )
        return all_group_index_to_group_name_mapping

    def group_names(self) -> List[str]:
        all_group_name_to_group_index_mapping: Dict[str, int] = (
            self.get_all_group_name_to_group_index_mapping()
        )
        all_group_names: List[str] = list(all_group_name_to_group_index_mapping.keys())
        return all_group_names



import datetime

import pytest

from great_expectations.core.util import (
    AzureUrl,
    DBFSPath,
    GCSUrl,
    S3Url,
    sniff_s3_compression,
    substitute_all_strftime_format_strings,
)


@pytest.mark.unit
def test_substitute_all_strftime_format_strings():
    now = datetime.datetime.now(datetime.timezone.utc)

    input_dict = {
        "month_no": "%m",
        "just_a_string": "Bloopy!",
        "string_with_month_word": "Today we are in the month %B!",
        "number": "90210",
        "escaped_percent": "'%%m' is the format string for month number",
        "inner_dict": {"day_word_full": "%A"},
        "list": ["a", 123, "%a"],
    }
    expected_output_dict = {
        "month_no": now.strftime("%m"),
        "just_a_string": "Bloopy!",
        "string_with_month_word": f"Today we are in the month {now.strftime('%B')}!",
        "number": "90210",
        "escaped_percent": "'%m' is the format string for month number",
        "inner_dict": {"day_word_full": now.strftime("%A")},
        "list": ["a", 123, now.strftime("%a")],
    }
    assert substitute_all_strftime_format_strings(input_dict) == expected_output_dict


@pytest.mark.parametrize(
    "url,expected",
    [
        ("s3://bucket/hello/world.csv.gz", "gz"),
        ("s3://bucket/hello/world.csv", "csv"),
        ("s3://bucket/hello/world.csv.gz?asd", "gz"),
        ("s3://bucket/hello/world?asd", None),
        ("<scheme>://<netloc>/<path>.csv;<params>?<query>#<fragment>", "csv"),
        ("<scheme>://<netloc>/<path>.;<params>?<query>#<fragment>", None),
    ],
)
@pytest.mark.unit
def test_s3_suffix(url, expected):
    _suffix = S3Url(url).suffix
    if expected is not None:
        assert _suffix == expected
    else:
        assert _suffix is None


@pytest.mark.parametrize(
    "url,expected",
    [
        ("s3://bucket/hello/world.csv.gz", "gzip"),
        ("s3://bucket/hello/world.csv", None),
        ("s3://bucket/hello/world.csv.xz", "xz"),
        ("s3://bucket/hello/world", None),
    ],
)
@pytest.mark.unit
def test_sniff_s3_compression(url, expected):
    assert sniff_s3_compression(S3Url(url)) == expected


@pytest.mark.unit
def test_azure_pandas_url():
    url = AzureUrl("my_account.blob.core.windows.net/my_container/my_blob")
    assert url.account_name == "my_account"
    assert url.container == "my_container"
    assert url.blob == "my_blob"


@pytest.mark.unit
def test_azure_pandas_url_with_https():
    url = AzureUrl("https://my_account.blob.core.windows.net/my_container/my_blob")
    assert url.account_name == "my_account"
    assert url.container == "my_container"
    assert url.blob == "my_blob"


@pytest.mark.unit
def test_azure_pandas_url_with_nested_blob():
    url = AzureUrl("my_account.blob.core.windows.net/my_container/a/b/c/d/e/my_blob")
    assert url.account_name == "my_account"
    assert url.container == "my_container"
    assert url.blob == "a/b/c/d/e/my_blob"


@pytest.mark.unit
def test_azure_pandas_url_with_special_chars():
    # Note that `url` conforms with the naming restrictions set by the Azure API
    # Azure naming restrictions: https://docs.microsoft.com/en-us/rest/api/storageservices/naming-and-referencing-containers--blobs--and-metadata
    url = AzureUrl("my_account.blob.core.windows.net/my-container_1.0/my-blob_`~!@#$%^&*()=+")
    assert url.account_name == "my_account"
    assert url.account_url == "my_account.blob.core.windows.net"
    assert url.container == "my-container_1.0"
    assert url.blob == "my-blob_`~!@#$%^&*()=+"


@pytest.mark.unit
def test_azure_spark_url():
    url = AzureUrl("my_container@my_account.blob.core.windows.net/my_blob")
    assert url.account_name == "my_account"
    assert url.account_url == "my_account.blob.core.windows.net"
    assert url.container == "my_container"
    assert url.blob == "my_blob"


@pytest.mark.unit
def test_azure_spark_url_with_wasbs():
    url = AzureUrl("wasbs://my_container@my_account.blob.core.windows.net/my_blob")
    assert url.account_name == "my_account"
    assert url.account_url == "my_account.blob.core.windows.net"
    assert url.container == "my_container"
    assert url.blob == "my_blob"


@pytest.mark.unit
def test_azure_spark_url_with_nested_blob():
    url = AzureUrl("my_container@my_account.blob.core.windows.net/a/b/c/d/e/my_blob")
    assert url.account_name == "my_account"
    assert url.account_url == "my_account.blob.core.windows.net"
    assert url.container == "my_container"
    assert url.blob == "a/b/c/d/e/my_blob"


@pytest.mark.unit
def test_azure_spark_url_with_special_chars():
    # Note that `url` conforms with the naming restrictions set by the Azure API
    # Azure naming restrictions: https://docs.microsoft.com/en-us/rest/api/storageservices/naming-and-referencing-containers--blobs--and-metadata
    url = AzureUrl("my-container_1.0@my_account.blob.core.windows.net/my-blob_`~!@#$%^&*()=+")
    assert url.account_name == "my_account"
    assert url.account_url == "my_account.blob.core.windows.net"
    assert url.container == "my-container_1.0"
    assert url.blob == "my-blob_`~!@#$%^&*()=+"


@pytest.mark.unit
def test_azure_url_with_invalid_url():
    with pytest.raises(AssertionError):
        AzureUrl("my_bucket/my_blob")


@pytest.mark.unit
def test_gcs_url():
    url = GCSUrl("gs://my_bucket/my_blob")
    assert url.bucket == "my_bucket"
    assert url.blob == "my_blob"


@pytest.mark.unit
def test_gcs_url_with_nested_blob():
    url = GCSUrl("gs://my_bucket/a/b/c/d/e/my_blob")
    assert url.bucket == "my_bucket"
    assert url.blob == "a/b/c/d/e/my_blob"


@pytest.mark.unit
def test_gcs_url_with_invalid_url():
    with pytest.raises(AssertionError):
        GCSUrl("my_bucket/my_blob")


@pytest.mark.unit
def test_gcs_url_with_special_chars():
    # Note that `url` conforms with the naming restrictions set by the GCS API
    # GCS bucket naming restrictions: https://cloud.google.com/storage/docs/naming-buckets
    # GCS blob naming restrictions: https://cloud.google.com/storage/docs/naming-objects
    url = GCSUrl("gs://my-bucket_1.0/my-blob_`~!@#$%^&*()=+")
    assert url.bucket == "my-bucket_1.0"
    assert url.blob == "my-blob_`~!@#$%^&*()=+"


@pytest.mark.parametrize(
    "input_path,expected_path",
    [
        # conversion
        ("/dbfs/some/path/myfile.csv", "dbfs:/some/path/myfile.csv"),
        ("/dbfs/myfile.csv", "dbfs:/myfile.csv"),
        ("/dbfs/myfolder", "dbfs:/myfolder"),
        ("/dbfs/myfolder/", "dbfs:/myfolder/"),
        ("/dbfs/my/nested/folder", "dbfs:/my/nested/folder"),
        ("/dbfs/my/nested/folder/", "dbfs:/my/nested/folder/"),
        # no conversion
        ("dbfs:/some/path/myfile.csv", "dbfs:/some/path/myfile.csv"),
        ("dbfs:/myfile.csv", "dbfs:/myfile.csv"),
        ("dbfs:/myfolder", "dbfs:/myfolder"),
        ("dbfs:/myfolder/", "dbfs:/myfolder/"),
        ("dbfs:/my/nested/folder", "dbfs:/my/nested/folder"),
        ("dbfs:/my/nested/folder/", "dbfs:/my/nested/folder/"),
        # trailing slash for root is optional in filesystem case, not in protocol case
        ("/dbfs/", "dbfs:/"),
        ("/dbfs", "dbfs:/"),
        ("dbfs:/", "dbfs:/"),
        ("dbfs:", "dbfs:/"),
    ],
)
@pytest.mark.unit
def test_dbfs_path_protocol_conversions(input_path, expected_path):
    observed_path = DBFSPath.convert_to_protocol_version(path=input_path)
    assert observed_path == expected_path


@pytest.mark.parametrize(
    "input_path,expected_path",
    [
        # conversion
        ("dbfs:/some/path/myfile.csv", "/dbfs/some/path/myfile.csv"),
        ("dbfs:/myfile.csv", "/dbfs/myfile.csv"),
        ("dbfs:/myfolder", "/dbfs/myfolder"),
        ("dbfs:/myfolder/", "/dbfs/myfolder/"),
        ("dbfs:/my/nested/folder", "/dbfs/my/nested/folder"),
        ("dbfs:/my/nested/folder/", "/dbfs/my/nested/folder/"),
        # no conversion
        ("/dbfs/some/path/myfile.csv", "/dbfs/some/path/myfile.csv"),
        ("/dbfs/myfile.csv", "/dbfs/myfile.csv"),
        ("/dbfs/myfolder", "/dbfs/myfolder"),
        ("/dbfs/myfolder/", "/dbfs/myfolder/"),
        ("/dbfs/my/nested/folder", "/dbfs/my/nested/folder"),
        ("/dbfs/my/nested/folder/", "/dbfs/my/nested/folder/"),
        # trailing slash for root is optional in filesystem case, not in protocol case
        ("/dbfs/", "/dbfs/"),
        ("/dbfs", "/dbfs"),
        ("dbfs:/", "/dbfs/"),
        ("dbfs:", "/dbfs"),
    ],
)
@pytest.mark.unit
def test_dbfs_path_file_semantics_conversions(input_path, expected_path):
    observed_path = DBFSPath.convert_to_file_semantics_version(path=input_path)
    assert observed_path == expected_path



from __future__ import annotations

from hashlib import md5


def tuple_to_hash(tuple_):
    return md5(str(tuple_).encode("utf-8")).hexdigest()


def kwargs_to_tuple(d):
    """Convert expectation configuration kwargs to a canonical tuple."""
    if isinstance(d, list):
        return tuple(kwargs_to_tuple(v) for v in sorted(d))
    elif isinstance(d, dict):
        return tuple(
            (k, kwargs_to_tuple(v))
            for k, v in sorted(d.items())
            if k not in ["result_format", "catch_exceptions", "meta"]
        )
    return d



from __future__ import annotations

import pathlib

import pytest

import great_expectations as gx
from great_expectations.data_context import get_context
from great_expectations.data_context.store.gx_cloud_store_backend import (
    GXCloudStoreBackend,
)
from great_expectations.data_context.store.inline_store_backend import (
    InlineStoreBackend,
)
from great_expectations.data_context.types.base import (
    DataContextConfig,
    GXCloudConfig,
)


@pytest.fixture
def pandas_enabled_datasource_config() -> dict:
    name = "my_pandas_datasource"
    class_name = "Datasource"
    execution_engine = {
        "class_name": "PandasExecutionEngine",
    }
    data_connectors = {
        "my_inferred_data_connector_name": {
            "class_name": "InferredAssetFilesystemDataConnector",
            "base_directory": "../data/",
            "default_regex": {
                "group_names": ["data_asset_name"],
                "pattern": "(.*)",
            },
        },
    }

    config = {
        "name": name,
        "class_name": class_name,
        "execution_engine": execution_engine,
        "data_connectors": data_connectors,
    }
    return config


@pytest.mark.unit
def test_data_context_instantiates_gx_cloud_store_backend_with_cloud_config(
    tmp_path: pathlib.Path,
    data_context_config_with_datasources: DataContextConfig,
    ge_cloud_config: GXCloudConfig,
) -> None:
    project_path = tmp_path / "my_data_context"
    project_path.mkdir()

    context = get_context(
        project_config=data_context_config_with_datasources,
        context_root_dir=str(project_path),
        cloud_base_url=ge_cloud_config.base_url,
        cloud_access_token=ge_cloud_config.access_token,
        cloud_organization_id=ge_cloud_config.organization_id,
        cloud_mode=True,
    )

    assert isinstance(context._datasource_store.store_backend, GXCloudStoreBackend)


@pytest.mark.filesystem
def test_data_context_instantiates_inline_store_backend_with_filesystem_config(
    tmp_path: pathlib.Path,
    data_context_config_with_datasources: DataContextConfig,
) -> None:
    project_path = tmp_path / "my_data_context"
    project_path.mkdir()

    context = get_context(
        project_config=data_context_config_with_datasources,
        context_root_dir=str(project_path),
        cloud_mode=False,
    )

    assert isinstance(context._datasource_store.store_backend, InlineStoreBackend)


@pytest.mark.unit
def test_list_datasources() -> None:
    context = gx.get_context(mode="ephemeral")

    context.data_sources.add_pandas("a")
    context.data_sources.add_pandas("b")

    assert len(context.list_datasources()) == 2


@pytest.mark.unit
def test_get_data_source() -> None:
    context = gx.get_context(mode="ephemeral")
    context.data_sources.add_pandas("a")
    ds = context.data_sources.get("a")
    assert ds.name == "a"


@pytest.mark.filesystem
def test_get_available_data_assets_names(empty_data_context) -> None:
    datasource_name = "my_fluent_pandas_datasource"
    datasource = empty_data_context.data_sources.add_pandas(datasource_name)
    asset_name = "test_data_frame"
    datasource.add_dataframe_asset(name=asset_name)

    assert len(empty_data_context.get_available_data_asset_names()) == 1

    data_asset_names = dict(
        empty_data_context.get_available_data_asset_names(datasource_names=datasource_name)
    )

    assert asset_name in data_asset_names[datasource_name]



from __future__ import annotations

import logging
from typing import Dict, List, cast

import pandas as pd
import pytest

from great_expectations.compatibility import sqlalchemy
from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.core import (
    ExpectationValidationResult,
)
from great_expectations.core.expectation_diagnostics.expectation_test_data_cases import (
    ExpectationTestCase,
    ExpectationTestDataCases,
    TestBackend,
    TestData,
)
from great_expectations.core.expectation_diagnostics.supporting_types import (
    ExpectationExecutionEngineDiagnostics,
)
from great_expectations.exceptions import GreatExpectationsError
from great_expectations.execution_engine import (
    ExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.expectation import (
    render_suite_parameter_string,
)
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.expectations.metrics.util import column_reflection_fallback
from great_expectations.render import RenderedStringTemplateContent
from great_expectations.self_check.util import (
    build_sa_validator_with_data,
    generate_expectation_tests,
    generate_test_table_name,
    should_we_generate_this_test,
)
from great_expectations.self_check.util import (
    build_test_backends_list as build_test_backends_list_v3,
)
from great_expectations.validator.metric_configuration import MetricConfiguration
from great_expectations.validator.validator import Validator

logger = logging.getLogger(__name__)


def get_table_columns_metric(
    execution_engine: ExecutionEngine,
) -> tuple[MetricConfiguration, dict]:
    resolved_metrics: dict = {}

    results: dict

    table_column_types_metric: MetricConfiguration = MetricConfiguration(
        metric_name="table.column_types",
        metric_domain_kwargs={},
        metric_value_kwargs={
            "include_nested": True,
        },
    )
    results = execution_engine.resolve_metrics(metrics_to_resolve=(table_column_types_metric,))
    resolved_metrics.update(results)

    table_columns_metric: MetricConfiguration = MetricConfiguration(
        metric_name="table.columns",
        metric_domain_kwargs={},
        metric_value_kwargs=None,
    )
    table_columns_metric.metric_dependencies = {  # type: ignore[assignment] # FIXME CoP
        "table.column_types": table_column_types_metric,
    }
    results = execution_engine.resolve_metrics(
        metrics_to_resolve=(table_columns_metric,), metrics=resolved_metrics
    )
    resolved_metrics.update(results)

    return table_columns_metric, resolved_metrics


@pytest.fixture(scope="module")
def expectation_and_runtime_configuration_with_suite_parameters():
    configuration = ExpectationConfiguration(
        type="expect_column_min_to_be_between",
        kwargs={
            "column": "live",
            "min_value": {"$PARAMETER": "MIN_VAL_PARAM"},
            "max_value": {"$PARAMETER": "MAX_VAL_PARAM"},
            "result_format": "SUMMARY",
        },
        meta={"BasicDatasetProfiler": {"confidence": "very low"}},
    )
    # runtime configuration with suite_parameters loaded
    runtime_configuration_with_eval = {
        "styling": {
            "default": {"classes": ["badge", "badge-secondary"]},
            "params": {"column": {"classes": ["badge", "badge-primary"]}},
        },
        "include_column_name": None,
        "suite_parameters": {"MIN_VAL_PARAM": 15, "MAX_VAL_PARAM": 20},
    }
    return configuration, runtime_configuration_with_eval


@pytest.mark.unit
def test_prescriptive_renderer_no_decorator(
    expectation_and_runtime_configuration_with_suite_parameters,
):
    (
        configuration,
        runtime_configuration_with_eval,
    ) = expectation_and_runtime_configuration_with_suite_parameters

    # noinspection PyShadowingNames
    def bare_bones_prescriptive_renderer(
        configuration=None,
        runtime_configuration=None,
    ):
        runtime_configuration = runtime_configuration or {}
        styling = runtime_configuration.get("styling")
        params = configuration.kwargs
        template_str = "$column minimum value must be greater than or equal to $min_value and less than or equal to $max_value"  # noqa: E501 # FIXME CoP
        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]

    res = bare_bones_prescriptive_renderer(
        configuration=configuration,
        runtime_configuration=runtime_configuration_with_eval,
    )
    assert len(res) == 1
    # string template should remain constant
    assert (
        res[0].string_template["template"]
        == "$column minimum value must be greater than or equal to $min_value and less than or equal to $max_value"  # noqa: E501 # FIXME CoP
    )

    # params should contain our suite parameters
    assert res[0].string_template["params"]["min_value"] == {"$PARAMETER": "MIN_VAL_PARAM"}
    assert res[0].string_template["params"]["max_value"] == {"$PARAMETER": "MAX_VAL_PARAM"}

    # full json dict comparison
    assert res[0].to_json_dict() == {
        "content_block_type": "string_template",
        "string_template": {
            "template": "$column minimum value must be greater than or equal to $min_value and less than or equal to $max_value",  # noqa: E501 # FIXME CoP
            "params": {
                "column": "live",
                "min_value": {"$PARAMETER": "MIN_VAL_PARAM"},
                "max_value": {"$PARAMETER": "MAX_VAL_PARAM"},
                "result_format": "SUMMARY",
            },
            "styling": {
                "default": {"classes": ["badge", "badge-secondary"]},
                "params": {"column": {"classes": ["badge", "badge-primary"]}},
            },
        },
    }


@pytest.mark.big
def test_prescriptive_renderer_with_decorator(
    expectation_and_runtime_configuration_with_suite_parameters,
):
    (
        configuration,
        runtime_configuration_with_eval,
    ) = expectation_and_runtime_configuration_with_suite_parameters

    # noinspection PyShadowingNames
    @render_suite_parameter_string
    def bare_bones_prescriptive_renderer(
        configuration=None,
        runtime_configuration=None,
    ):
        runtime_configuration = runtime_configuration or {}
        styling = runtime_configuration.get("styling")
        params = configuration.kwargs
        template_str = "$column minimum value must be greater than or equal to $min_value and less than or equal to $max_value"  # noqa: E501 # FIXME CoP
        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]

    res = bare_bones_prescriptive_renderer(
        configuration=configuration,
        runtime_configuration=runtime_configuration_with_eval,
    )
    assert len(res) == 3

    # string template should remain constant
    assert (
        res[0].string_template["template"]
        == "$column minimum value must be greater than or equal to $min_value and less than or equal to $max_value"  # noqa: E501 # FIXME CoP
    )

    # params should contain our suite parameters
    assert res[0].string_template["params"]["min_value"] == {"$PARAMETER": "MIN_VAL_PARAM"}
    assert res[0].string_template["params"]["max_value"] == {"$PARAMETER": "MAX_VAL_PARAM"}
    assert res[0].to_json_dict() == {
        "content_block_type": "string_template",
        "string_template": {
            "template": "$column minimum value must be greater than or equal to $min_value and less than or equal to $max_value",  # noqa: E501 # FIXME CoP
            "params": {
                "column": "live",
                "min_value": {"$PARAMETER": "MIN_VAL_PARAM"},
                "max_value": {"$PARAMETER": "MAX_VAL_PARAM"},
                "result_format": "SUMMARY",
            },
            "styling": {
                "default": {"classes": ["badge", "badge-secondary"]},
                "params": {"column": {"classes": ["badge", "badge-primary"]}},
            },
        },
    }

    assert (
        res[1].string_template["template"]
        == "\n - $eval_param = $eval_param_value (at time of validation)."
    )
    # params should contain our suite parameters
    assert res[1].string_template["params"]["eval_param"] == "MIN_VAL_PARAM"
    assert res[1].string_template["params"]["eval_param_value"] == 15
    assert res[1].to_json_dict() == {
        "content_block_type": "string_template",
        "string_template": {
            "template": "\n - $eval_param = $eval_param_value (at time of validation).",
            "params": {"eval_param": "MIN_VAL_PARAM", "eval_param_value": 15},
            "styling": {
                "default": {"classes": ["badge", "badge-secondary"]},
                "params": {"column": {"classes": ["badge", "badge-primary"]}},
            },
        },
    }

    assert (
        res[2].string_template["template"]
        == "\n - $eval_param = $eval_param_value (at time of validation)."
    )
    # params should contain our suite parameters
    assert res[2].string_template["params"]["eval_param"] == "MAX_VAL_PARAM"
    assert res[2].string_template["params"]["eval_param_value"] == 20
    assert res[2].to_json_dict() == {
        "content_block_type": "string_template",
        "string_template": {
            "template": "\n - $eval_param = $eval_param_value (at time of validation).",
            "params": {"eval_param": "MAX_VAL_PARAM", "eval_param_value": 20},
            "styling": {
                "default": {"classes": ["badge", "badge-secondary"]},
                "params": {"column": {"classes": ["badge", "badge-primary"]}},
            },
        },
    }

    # with no runtime_configuration, throw an error
    with pytest.raises(GreatExpectationsError):
        # noinspection PyUnusedLocal
        res = bare_bones_prescriptive_renderer(
            configuration=configuration, runtime_configuration={}
        )

    # configuration should always be of ExpectationConfiguration-type
    with pytest.raises(AttributeError):
        # noinspection PyUnusedLocal,PyTypeChecker
        res = bare_bones_prescriptive_renderer(configuration={}, runtime_configuration={})

    # extra suite parameters will not have an effect
    runtime_configuration_with_extra = {
        "styling": {
            "default": {"classes": ["badge", "badge-secondary"]},
            "params": {"column": {"classes": ["badge", "badge-primary"]}},
        },
        "include_column_name": None,
        "suite_parameters": {
            "MIN_VAL_PARAM": 15,
            "MAX_VAL_PARAM": 20,
            "IAMEXTRA": "EXTRA",
        },
    }

    res = bare_bones_prescriptive_renderer(
        configuration=configuration,
        runtime_configuration=runtime_configuration_with_extra,
    )
    assert len(res) == 3

    # missing suite_parameters will not render (MAX_VAL_PARAM is missing)
    runtime_configuration_with_missing = {
        "styling": {
            "default": {"classes": ["badge", "badge-secondary"]},
            "params": {"column": {"classes": ["badge", "badge-primary"]}},
        },
        "include_column_name": None,
        "suite_parameters": {"MIN_VAL_PARAM": 15},
    }
    res = bare_bones_prescriptive_renderer(
        configuration=configuration,
        runtime_configuration=runtime_configuration_with_missing,
    )
    assert len(res) == 2


# noinspection PyUnusedLocal
@pytest.mark.all_backends
def test_table_column_reflection_fallback(test_backends, sa):
    include_sqlalchemy: bool = "sqlite" in test_backends
    include_postgresql: bool = "postgresql" in test_backends
    include_mysql: bool = "mysql" in test_backends
    include_mssql: bool = "mssql" in test_backends
    include_bigquery: bool = "bigquery" in test_backends
    include_trino: bool = "trino" in test_backends
    include_clickhouse: bool = "clickhouse" in test_backends

    if not sa.create_engine:
        pytest.skip("Unable to import sqlalchemy.create_engine() -- skipping.")

    test_backend_names: List[str] = build_test_backends_list_v3(
        include_pandas=False,
        include_spark=False,
        include_sqlalchemy=include_sqlalchemy,
        include_postgresql=include_postgresql,
        include_mysql=include_mysql,
        include_mssql=include_mssql,
        include_bigquery=include_bigquery,
        include_trino=include_trino,
        include_clickhouse=include_clickhouse,
    )

    df: pd.DataFrame = pd.DataFrame(
        {
            "name": ["Frank", "Steve", "Jane", "Frank", "Michael"],
            "age": [16, 21, 38, 22, 10],
            "pet": ["fish", "python", "cat", "python", "frog"],
        }
    )

    validators_config: Dict[str, Validator] = {}
    validator: Validator
    backend_name: str
    table_name: str
    for backend_name in test_backend_names:
        if backend_name in ["sqlite", "postgresql", "mysql", "mssql", "trino"]:
            table_name = generate_test_table_name()
            validator = build_sa_validator_with_data(
                df=df,
                sa_engine_name=backend_name,
                schemas=None,
                caching=True,
                table_name=table_name,
                sqlite_db_path=None,
            )
            if validator is not None:
                validators_config[table_name] = validator

    engine: sqlalchemy.Engine

    metrics: dict = {}

    table_columns_metric: MetricConfiguration
    results: dict

    reflected_columns_list: List[Dict[str, str]]
    reflected_column_config: Dict[str, str]
    column_name: str

    validation_result: ExpectationValidationResult

    sqlalchemy_engine: SqlAlchemyExecutionEngine

    for table_name, validator in validators_config.items():
        table_columns_metric, results = get_table_columns_metric(
            execution_engine=validator.execution_engine
        )
        metrics.update(results)
        assert set(metrics[table_columns_metric.id]) == {"name", "age", "pet"}
        selectable: sqlalchemy.Select = sa.Table(
            table_name,
            sa.MetaData(),
            schema=None,
        )
        sqlalchemy_engine = cast("SqlAlchemyExecutionEngine", validator.execution_engine)
        reflected_columns_list = column_reflection_fallback(
            selectable=selectable,
            dialect=sqlalchemy_engine.engine.dialect,
            sqlalchemy_engine=sqlalchemy_engine.engine,
        )
        for column_name in [
            reflected_column_config["name"] for reflected_column_config in reflected_columns_list
        ]:
            validation_result = validator.expect_column_to_exist(column=column_name)
            assert validation_result.success

    if validators_config:
        validator = list(validators_config.values())[0]

        validation_result = validator.expect_column_mean_to_be_between(column="age", min_value=10)
        assert validation_result.success

        validation_result = validator.expect_table_row_count_to_equal(value=5)
        assert validation_result.success

        validation_result = validator.expect_table_row_count_to_equal(value=3)
        assert not validation_result.success


@pytest.mark.skip(
    reason="Timeout of 30 seconds reached trying to connect to localhost:8088 (trino port)"
)
@pytest.mark.skipif(
    sa is None,
    reason="sqlalchemy is not installed",
)
@pytest.mark.all_backends
def test__generate_expectation_tests():
    expectation_type = "whatever"
    data = TestData(stuff=[1, 2, 3, 4, 5])
    test_case = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=[],
        only_for=[],
    )
    test_data_cases = [
        ExpectationTestDataCases(
            data=data,
            tests=[test_case],
        )
    ]
    engines = ExpectationExecutionEngineDiagnostics(
        PandasExecutionEngine=True,
        SqlAlchemyExecutionEngine=True,
        SparkDFExecutionEngine=False,
    )

    results = generate_expectation_tests(
        expectation_type=expectation_type,
        test_data_cases=test_data_cases,
        execution_engine_diagnostics=engines,
        raise_exceptions_for_backends=False,
    )
    backends_to_use = [r["backend"] for r in results]

    # If another SQL backend is available wherever this test is being run, it will
    # be included (i.e. postgresql)
    assert "pandas" in backends_to_use
    # assert "sqlite" in backends_to_use
    assert "spark" not in backends_to_use


@pytest.mark.unit
def test__TestBackend__bad_backends():
    with pytest.raises(AssertionError):
        TestBackend(
            backend="dogs",
            dialects=None,
        )


@pytest.mark.unit
def test__TestBackend__bad_dialects():
    with pytest.raises(AssertionError):
        TestBackend(
            backend="sqlalchemy",
            dialects=None,
        )

    with pytest.raises(AssertionError):
        TestBackend(
            backend="sqlalchemy",
            dialects=[],
        )

    with pytest.raises(AssertionError):
        TestBackend(
            backend="sqlalchemy",
            dialects=["postgresql", "mysql", "ramen"],
        )

    with pytest.raises(AssertionError):
        TestBackend(
            backend="spark",
            dialects=["sqlite"],
        )

    with pytest.raises(AssertionError):
        TestBackend(
            backend="pandas",
            dialects=["sqlite"],
        )

    with pytest.raises(AssertionError):
        TestBackend(
            backend="sqlalchemy",
            dialects="sqlite",
        )

    TestBackend(
        backend="sqlalchemy",
        dialects=["sqlite"],
    )


@pytest.mark.unit
def test__TestBackend__good_backends_and_dialects():
    tb1 = TestBackend(  # noqa: F841 # FIXME CoP
        backend="pandas",
        dialects=None,
    )

    tb2 = TestBackend(  # noqa: F841 # FIXME CoP
        backend="spark",
        dialects=None,
    )

    tb3 = TestBackend(  # noqa: F841 # FIXME CoP
        backend="sqlalchemy",
        dialects=["sqlite", "postgresql", "mysql"],
    )


@pytest.mark.unit
def test__should_we_generate_this_test__obvious():
    test_case = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=[],
        only_for=[],
    )
    backend = "spark"

    assert should_we_generate_this_test(backend, test_case) is True

    test_case2 = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=[],
        only_for=["postgresql"],
    )
    backend2 = "sqlite"

    assert should_we_generate_this_test(backend2, test_case2) is False

    test_case3 = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=[],
        only_for=["postgresql", "sqlite"],
    )
    backend3 = "sqlite"

    assert should_we_generate_this_test(backend3, test_case3) is True

    test_case4 = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=["sqlite"],
        only_for=[],
    )
    backend4 = "sqlite"

    assert should_we_generate_this_test(backend4, test_case4) is False

    test_case5 = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=["postgresql", "mssql"],
        only_for=["pandas", "sqlite"],
    )
    backend5 = "pandas"

    assert should_we_generate_this_test(backend5, test_case5) is True


# this is really a series of unit tests
@pytest.mark.unit
def test__should_we_generate_this_test__sqlalchemy():
    test_case = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=[],
        only_for=["sqlalchemy"],
    )
    backend = "mysql"

    assert should_we_generate_this_test(backend, test_case) is True

    test_case2 = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=[],
        only_for=["sqlalchemy"],
    )
    backend2 = "postgresql"

    assert should_we_generate_this_test(backend2, test_case2) is True

    test_case3 = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=["mysql"],
        only_for=["sqlalchemy"],
    )
    backend3 = "mysql"

    assert should_we_generate_this_test(backend3, test_case3) is False

    test_case4 = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=["sqlalchemy"],
        only_for=[],
    )
    backend4 = "sqlite"

    assert should_we_generate_this_test(backend4, test_case4) is False

    test_case5 = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=["sqlalchemy"],
        only_for=[],
    )
    backend5 = "spark"

    assert should_we_generate_this_test(backend5, test_case5) is True


@pytest.mark.unit
def test__should_we_generate_this_test__pandas():
    """
    Our CI/CD runs tests against pandas versions 0.23.4, 0.25.3, and latest (1.x currently)

    See: azure-pipelines.yml in project root
    """
    major, minor, *_ = pd.__version__.split(".")

    test_case = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=[],
        only_for=[],
    )
    backend = "pandas"

    assert should_we_generate_this_test(backend, test_case) is True

    test_case2 = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=[],
        only_for=["pandas", "spark"],
    )
    backend2 = "pandas"

    assert should_we_generate_this_test(backend2, test_case2) is True

    test_case3 = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=[],
        only_for=["pandas>=024"],
    )
    backend3 = "pandas"

    expected3 = False
    if (major == "0" and int(minor) >= 24) or int(major) >= 1:
        expected3 = True

    assert should_we_generate_this_test(backend3, test_case3) == expected3

    test_case4 = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=[],
        only_for=["pandas_023"],
    )
    backend4 = "pandas"

    expected4 = False
    if major == "0" and minor == "23":
        expected4 = True

    assert should_we_generate_this_test(backend4, test_case4) == expected4

    test_case5 = ExpectationTestCase(
        title="",
        input={},
        output={},
        exact_match_out=False,
        include_in_gallery=False,
        suppress_test_for=["pandas"],
        only_for=[],
    )
    backend5 = "pandas"

    assert should_we_generate_this_test(backend5, test_case5) is False



from __future__ import annotations

from typing import Any, ClassVar, Dict

from great_expectations.execution_engine import (
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.metric_provider import metric_value
from great_expectations.expectations.metrics.query_metric_provider import (
    InvalidParameterTypeError,
    MissingParameterError,
    QueryMetricProvider,
)


class QueryDataSourceTable(QueryMetricProvider):
    metric_name = "query.data_source_table"
    value_keys = ("query", "data_source_name")
    data_source_name_param_name: ClassVar[str] = "data_source_name"

    @classmethod
    def _get_data_source_name_from_metric_value_kwargs(cls, metric_value_kwargs: dict) -> str:
        data_source_name_param = cls.data_source_name_param_name
        data_source_name: str | None = metric_value_kwargs.get(
            data_source_name_param
        ) or cls.default_kwarg_values.get(data_source_name_param)
        if not data_source_name:
            raise MissingParameterError(data_source_name_param)
        if not isinstance(data_source_name, str):
            raise InvalidParameterTypeError(data_source_name_param, str)

        return data_source_name

    @metric_value(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(
        cls,
        execution_engine: SqlAlchemyExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ) -> list[dict]:
        from great_expectations.data_context import project_manager

        data_source_name = cls._get_data_source_name_from_metric_value_kwargs(metric_value_kwargs)
        # we're querying data from another data source,
        # so we need to use the execution engine of that data source.
        data_source = project_manager.get_datasources()[data_source_name]
        other_execution_engine = data_source.get_execution_engine()
        query = cls._get_query_from_metric_value_kwargs(metric_value_kwargs)
        return cls._get_sqlalchemy_records_from_substituted_batch_subquery(
            substituted_batch_subquery=query,
            execution_engine=other_execution_engine,
        )



from __future__ import annotations

from typing import Any, Callable, TypeVar

try:
    # default to the typing_extensions version if available as it contains bug fixes & improvements
    from typing_extensions import Annotated
except ImportError:
    from typing import Annotated  # type: ignore[assignment] # FIXME CoP

try:
    from typing_extensions import override
except ImportError:
    F = TypeVar("F", bound=Callable[..., Any])

    def override(__arg: F, /) -> F:
        return __arg


__all__ = ["Annotated", "override"]



import os
from typing import Tuple

from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.types.base import DataContextConfig
from great_expectations.data_context.types.refs import GXCloudResourceRef


def make_retrieve_data_context_config_from_cloud(data_dir):
    def retrieve_data_context_config_from_cloud(cls, *args, **kwargs):
        return DataContextConfig(**_cloud_config(data_dir))

    return retrieve_data_context_config_from_cloud


def make_store_get(data_file_name, data_dir, with_slack):
    def store_get(self, key):
        # key is a 3-tuple with the form
        # (GXCloudRESTResource, cloud_id as a string uuid, name as string)
        # For example:
        # (<GXCloudRESTResource.CHECKPOINT: 'checkpoint'>, '731dc2a5-45d8-4827-9118-39b77c5cd413', 'my_checkpoint')  # noqa: E501 # FIXME CoP
        type_ = key[0]
        if type_ == GXCloudRESTResource.CHECKPOINT:
            return {"data": _checkpoint_config(data_file_name, with_slack)}
        elif type_ == GXCloudRESTResource.EXPECTATION_SUITE:
            return {"data": _expectation_suite()}
        elif type_ == GXCloudRESTResource.DATASOURCE:
            return {"data": _datasource(data_dir)}
        return None

    return store_get


def store_set(self, key, value, **kwargs):
    base_url = os.environ["GX_CLOUD_BASE_URL"]
    org_id = os.environ["GX_CLOUD_ORGANIZATION_ID"]
    if not base_url.endswith("/"):
        base_url += "/"
    url = base_url + "organizations/" + org_id + "/validation-results"
    # This is incomplete but has the necessary info for the current tests.
    validation_result_json = {
        "data": {
            "attributes": {
                "created_by_id": "934e0898-6a5c-4ffd-9125-89381a46d191",
                "organization_id": org_id,
                "validation_result": {
                    "display_url": f"{base_url}{org_id}/?validationResultId=2e13ecc3-eaaa-444b-b30d-2f616f80ae35",  # noqa: E501 # FIXME CoP
                },
            }
        }
    }
    return GXCloudResourceRef(
        resource_type=GXCloudRESTResource.VALIDATION_RESULT,
        id="2e13ecc3-eaaa-444b-b30d-2f616f80ae35",
        url=url,
        response_json=validation_result_json,
    )


def list_keys(self, prefix: Tuple = ()):
    type_ = self._ge_cloud_resource_type
    if type_ == GXCloudRESTResource.EXPECTATION_SUITE:
        return [
            (
                GXCloudRESTResource.EXPECTATION_SUITE,
                "1212e79d-f751-4c6e-921d-26de2b1db174",
                "suite_name",
            ),
        ]
    elif type_ == GXCloudRESTResource.DATASOURCE:
        return [
            (
                GXCloudRESTResource.DATASOURCE,
                "eb0c729d-9457-43a0-8b40-6ec6c79c0fef",
                "taxi_datasource",
            ),
        ]
    raise NotImplementedError(f"{type_} is not supported in this mock")


class CallCounter:
    def __init__(self):
        self.count = 0

    def increment(self):
        self.count += 1


def make_send_slack_notifications(counter: CallCounter):
    def send_slack_notification(*args, **kwargs):
        counter.increment()

    return send_slack_notification


def _cloud_config(data_dir):
    return {
        "data_context_id": "6a52bdfa-e182-455b-a825-e69f076e67d6",
        "analytics_enabled": True,
        "checkpoint_store_name": "default_checkpoint_store",
        "config_variables_file_path": "uncommitted/config_variables.yml",
        "config_version": 3.0,
        "data_docs_sites": {},
        "expectations_store_name": "default_expectations_store",
        "plugins_directory": "plugins/",
        "progress_bars": {
            "globally": False,
            "metric_calculations": False,
            "profilers": False,
        },
        "stores": {
            "default_checkpoint_store": {
                "class_name": "CheckpointStore",
                "store_backend": {
                    "class_name": "GXCloudStoreBackend",
                    "ge_cloud_base_url": os.environ["GX_CLOUD_BASE_URL"],
                    "ge_cloud_credentials": {
                        "access_token": os.environ["GX_CLOUD_ACCESS_TOKEN"],
                        "organization_id": os.environ["GX_CLOUD_ORGANIZATION_ID"],
                    },
                    "ge_cloud_resource_type": "checkpoint",
                    "suppress_store_backend_id": True,
                },
            },
            "default_expectations_store": {
                "class_name": "ExpectationsStore",
                "store_backend": {
                    "class_name": "GXCloudStoreBackend",
                    "ge_cloud_base_url": os.environ["GX_CLOUD_BASE_URL"],
                    "ge_cloud_credentials": {
                        "access_token": os.environ["GX_CLOUD_ACCESS_TOKEN"],
                        "organization_id": os.environ["GX_CLOUD_ORGANIZATION_ID"],
                    },
                    "ge_cloud_resource_type": "expectation_suite",
                    "suppress_store_backend_id": True,
                },
            },
            "default_validation_results_store": {
                "class_name": "ValidationResultsStore",
                "store_backend": {
                    "class_name": "GXCloudStoreBackend",
                    "ge_cloud_base_url": os.environ["GX_CLOUD_BASE_URL"],
                    "ge_cloud_credentials": {
                        "access_token": os.environ["GX_CLOUD_ACCESS_TOKEN"],
                        "organization_id": os.environ["GX_CLOUD_ORGANIZATION_ID"],
                    },
                    "ge_cloud_resource_type": "validation_result",
                    "suppress_store_backend_id": True,
                },
            },
            "expectations_store": {
                "class_name": "ExpectationsStore",
                "store_backend": {
                    "base_directory": "expectations/",
                    "class_name": "TupleFilesystemStoreBackend",
                },
            },
        },
        "validation_results_store_name": "default_validation_results_store",
    }


def _checkpoint_config(data_file_name, with_slack):
    action_list = [
        {
            "action": {"class_name": "StoreValidationResultAction"},
            "name": "store_validation_result",
        },
    ]
    if with_slack:
        action_list.append(
            {
                "name": "send_slack_notification_on_validation_result",
                "action": {
                    "class_name": "SlackNotificationAction",
                    "slack_webhook": "https://hooks.slack.com/services/11111111111/22222222222/333333333333333333333333",
                    "notify_on": "all",
                    "renderer": {
                        "module_name": "great_expectations.render.renderer.slack_renderer",
                        "class_name": "SlackRenderer",
                    },
                },
            },
        )

    return [
        {
            "attributes": {
                "checkpoint_config": {
                    "action_list": action_list,
                    "batch_request": {},
                    "default_validation_id": "51c303bd-2396-4f04-b567-79f746b09173",
                    "suite_parameters": {},
                    "expectation_suite_id": None,
                    "expectation_suite_name": None,
                    "id": "731dc2a5-45d8-4827-9118-39b77c5cd413",
                    "name": "my_checkpoint",
                    "runtime_configuration": {},
                    "validations": [
                        {
                            "batch_request": {
                                "data_asset_name": data_file_name,
                                "data_connector_name": "taxi_data_connector",
                                "data_connector_query": {"index": -1},
                                "datasource_name": "taxi_datasource",
                            },
                            "expectation_suite_id": "d1ff9854-ac5f-45ae-9c88-c3f4323432c1",
                            "expectation_suite_name": "taxi_demo_suite",
                            "id": "2e13ecc3-eaaa-444b-b30d-2f616f80ae35",
                        }
                    ],
                },
                "created_by_id": "934e0898-6a5c-4ffd-9125-89381a46d191",
                "default_validation_id": "51c303bd-2396-4f04-b567-79f746b09173",
                "id": "731dc2a5-45d8-4827-9118-39b77c5cd413",
                "name": "my_checkpoint",
                "organization_id": os.environ["GX_CLOUD_ORGANIZATION_ID"],
            },
            "id": "731dc2a5-45d8-4827-9118-39b77c5cd413",
            "type": "checkpoint",
        }
    ]


def _expectation_suite():
    return [
        {
            "created_by_id": "934e0898-6a5c-4ffd-9125-89381a46d191",
            "organization_id": os.environ["GX_CLOUD_ORGANIZATION_ID"],
            "name": "single-snippet-suite-2",
            "expectations": [
                {
                    "expectation_type": "expect_column_to_exist",
                    "id": "0571b111-a1d0-4f1a-9c83-6704887de635",
                    "kwargs": {"column": "passenger_count"},
                    "meta": {},
                }
            ],
            "id": "1212e79d-f751-4c6e-921d-26de2b1db174",
            "meta": {"great_expectations_version": "0.15.43"},
        }
    ]


def _datasource(data_dir):
    return [
        {
            "attributes": {
                "created_by_id": "934e0898-6a5c-4ffd-9125-89381a46d191",
                "organization_id": os.environ["GX_CLOUD_ORGANIZATION_ID"],
                "datasource_config": {
                    "class_name": "Datasource",
                    "data_connectors": {
                        "default_runtime_data_connector_name": {
                            "assets": {
                                "taxi_data": {
                                    "batch_identifiers": ["runtime_batch_identifier_name"],
                                    "class_name": "Asset",
                                    "module_name": "great_expectations.datasource.data_connector.asset",  # noqa: E501 # FIXME CoP
                                }
                            },
                            "class_name": "RuntimeDataConnector",
                            "id": "e0af346c-32ea-44e6-8908-b559c4162a70",
                            "module_name": "great_expectations.datasource.data_connector",
                            "name": "default_runtime_data_connector_name",
                        },
                        "taxi_data_connector": {
                            "base_directory": str(data_dir),
                            "class_name": "InferredAssetFilesystemDataConnector",
                            "default_regex": {
                                "group_names": ["data_asset_name"],
                                "pattern": "(.*)",
                            },
                            "id": "997a7842-195b-4374-a71b-e52f192068d1",
                            "module_name": "great_expectations.datasource.data_connector",
                            "name": "taxi_data_connector",
                        },
                    },
                    "execution_engine": {
                        "class_name": "PandasExecutionEngine",
                        "module_name": "great_expectations.execution_engine",
                    },
                    "id": "eb0c729d-9457-43a0-8b40-6ec6c79c0fef",
                    "module_name": "great_expectations.datasource",
                    "name": "taxi_datasource",
                },
            },
            "id": "eb0c729d-9457-43a0-8b40-6ec6c79c0fef",
            "type": "datasource",
        }
    ]



from datetime import datetime

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    DATA_SOURCES_THAT_SUPPORT_DATE_COMPARISONS,
    JUST_PANDAS_DATA_SOURCES,
)

COL_NAME = "my_col"

ONES_AND_TWOS = pd.DataFrame({COL_NAME: [1, 2, 2, 2]})


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=ONES_AND_TWOS)
def test_success_complete_results(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnDistinctValuesToEqualSet(column=COL_NAME, value_set=[1, 2])
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {
        "details": {
            "value_counts": [
                {"value": 1, "count": 1},
                {"value": 2, "count": 3},
            ]
        },
        "observed_value": [1, 2],
    }


@parameterize_batch_for_data_sources(
    data_source_configs=ALL_DATA_SOURCES,
    data=pd.DataFrame({COL_NAME: ["foo", "bar"]}),
)
def test_strings(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnDistinctValuesToEqualSet(
        column=COL_NAME, value_set=["foo", "bar"]
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=DATA_SOURCES_THAT_SUPPORT_DATE_COMPARISONS,
    data=pd.DataFrame({COL_NAME: [datetime(2024, 11, 19).date(), datetime(2024, 11, 20).date()]}),  # noqa: DTZ001 # FIXME CoP
)
def test_dates(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnDistinctValuesToEqualSet(
        column=COL_NAME,
        value_set=[datetime(2024, 11, 19).date(), datetime(2024, 11, 20).date()],  # noqa: DTZ001 # FIXME CoP
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=DATA_SOURCES_THAT_SUPPORT_DATE_COMPARISONS,
    data=pd.DataFrame({COL_NAME: [datetime(2024, 11, 19).date(), datetime(2024, 11, 20).date()]}),  # noqa: DTZ001 # FIXME CoP
)
def test_dates_with_str_value_set(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnDistinctValuesToEqualSet(
        column=COL_NAME,
        value_set=[str(datetime(2024, 11, 19).date()), str(datetime(2024, 11, 20).date())],  # noqa: DTZ001 # FIXME CoP
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES, data=pd.DataFrame({COL_NAME: [1, 2, None]})
)
def test_ignores_nulls(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnDistinctValuesToEqualSet(column=COL_NAME, value_set=[1, 2])
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize("value_set", [[1], [1, 4], [1, 2, 3]])
@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES, data=ONES_AND_TWOS
)
def test_fails_if_data_is_not_equal(batch_for_datasource: Batch, value_set: list[int]) -> None:
    expectation = gxe.ExpectColumnDistinctValuesToEqualSet(column=COL_NAME, value_set=value_set)
    result = batch_for_datasource.validate(expectation)
    assert not result.success



import uuid
from unittest.mock import MagicMock, Mock  # noqa: TID251 # FIXME CoP

import pytest

from great_expectations.data_context import CloudDataContext
from great_expectations.datasource.fluent import BatchRequest
from great_expectations.experimental.metric_repository.batch_inspector import (
    BatchInspector,
)
from great_expectations.experimental.metric_repository.metric_retriever import (
    MetricRetriever,
)
from great_expectations.experimental.metric_repository.metrics import (
    MetricRun,
    MetricTypes,
    TableMetric,
)

pytestmark = pytest.mark.unit


# compute_metric_list_run tests
def test_compute_metric_list_run_with_no_metric_retrievers(mocker):
    mock_context = Mock(spec=CloudDataContext)
    batch_inspector = BatchInspector(context=mock_context, metric_retrievers=[])
    mock_batch_request = Mock(spec=BatchRequest)

    data_asset_id = uuid.uuid4()

    metric_run = batch_inspector.compute_metric_list_run(
        data_asset_id=data_asset_id, batch_request=mock_batch_request, metric_list=[]
    )
    assert metric_run == MetricRun(data_asset_id=data_asset_id, metrics=[])


def test_compute_metric_list_run_calls_metric_retrievers():
    mock_context = Mock(spec=CloudDataContext)
    mock_metric_retriever = MagicMock(spec=MetricRetriever)
    batch_inspector = BatchInspector(
        context=mock_context, metric_retrievers=[mock_metric_retriever]
    )
    mock_batch_request = Mock(spec=BatchRequest)

    data_asset_id = uuid.uuid4()

    metric_list = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
        MetricTypes.COLUMN_MIN,
        MetricTypes.COLUMN_MAX,
        MetricTypes.COLUMN_MEAN,
        MetricTypes.COLUMN_MEDIAN,
        MetricTypes.COLUMN_NULL_COUNT,
    ]

    batch_inspector.compute_metric_list_run(
        data_asset_id=data_asset_id,
        batch_request=mock_batch_request,
        metric_list=metric_list,
    )

    assert mock_metric_retriever.get_metrics.call_count == 1

    mock_metric_retriever.get_metrics.assert_called_once_with(
        batch_request=mock_batch_request, metric_list=metric_list
    )


def test_compute_metric_list_run_returns_metric_run():
    mock_context = Mock(spec=CloudDataContext)
    mock_metric_retriever = MagicMock(spec=MetricRetriever)

    mock_metric = Mock(spec=TableMetric)
    mock_metric_retriever.get_metrics.return_value = [mock_metric]

    batch_inspector = BatchInspector(
        context=mock_context, metric_retrievers=[mock_metric_retriever]
    )
    mock_batch_request = Mock(spec=BatchRequest)

    data_asset_id = uuid.uuid4()

    metric_run = batch_inspector.compute_metric_list_run(
        data_asset_id=data_asset_id,
        batch_request=mock_batch_request,
        metric_list=[
            MetricTypes.TABLE_ROW_COUNT,
            MetricTypes.TABLE_COLUMNS,
            MetricTypes.TABLE_COLUMN_TYPES,
            MetricTypes.COLUMN_MIN,
            MetricTypes.COLUMN_MAX,
            MetricTypes.COLUMN_MEAN,
            MetricTypes.COLUMN_MEDIAN,
            MetricTypes.COLUMN_NULL_COUNT,
        ],
    )

    assert metric_run == MetricRun(
        data_asset_id=data_asset_id,
        metrics=[mock_metric],
    )


# compute_metric_run tests.
# Will eventually go away once compute_metric_list_run is fully implemented.


def test_compute_metric_run_with_no_metric_retrievers():
    mock_context = Mock(spec=CloudDataContext)
    batch_inspector = BatchInspector(context=mock_context, metric_retrievers=[])
    mock_batch_request = Mock(spec=BatchRequest)

    data_asset_id = uuid.uuid4()

    metric_run = batch_inspector.compute_metric_run(
        data_asset_id=data_asset_id, batch_request=mock_batch_request
    )
    assert metric_run == MetricRun(data_asset_id=data_asset_id, metrics=[])


def test_compute_metric_run_calls_metric_retrievers():
    mock_context = Mock(spec=CloudDataContext)
    mock_metric_retriever = MagicMock(spec=MetricRetriever)
    batch_inspector = BatchInspector(
        context=mock_context, metric_retrievers=[mock_metric_retriever]
    )
    mock_batch_request = Mock(spec=BatchRequest)

    data_asset_id = uuid.uuid4()

    batch_inspector.compute_metric_run(
        data_asset_id=data_asset_id, batch_request=mock_batch_request
    )

    assert mock_metric_retriever.get_metrics.call_count == 1

    mock_metric_retriever.get_metrics.assert_called_once_with(batch_request=mock_batch_request)


def test_compute_metric_run_returns_metric_run():
    mock_context = Mock(spec=CloudDataContext)
    mock_metric_retriever = MagicMock(spec=MetricRetriever)

    mock_metric = Mock(spec=TableMetric)
    mock_metric_retriever.get_metrics.return_value = [mock_metric]

    batch_inspector = BatchInspector(
        context=mock_context, metric_retrievers=[mock_metric_retriever]
    )
    mock_batch_request = Mock(spec=BatchRequest)

    data_asset_id = uuid.uuid4()

    metric_run = batch_inspector.compute_metric_run(
        data_asset_id=data_asset_id, batch_request=mock_batch_request
    )

    assert metric_run == MetricRun(
        data_asset_id=data_asset_id,
        metrics=[mock_metric],
    )



from __future__ import annotations

import itertools
import logging
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Sequence, Union

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility.pydantic import StrictInt, StrictStr
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.id_dict import IDDict

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

    from great_expectations.core.batch import LegacyBatchDefinition

logger = logging.getLogger(__name__)


class SliceValidator:
    """
    A custom slice class which has implemented:
      - __get_validators__ for type validation
      - __modify_schemas__ to provide custom json schema
    """

    def __init__(self, slice_validator: slice):
        self._slice = slice_validator
        self.start = self._slice.start
        self.stop = self._slice.stop
        self.step = self._slice.step

    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        # __modify_schema__ should mutate the dict it receives in place,
        # the returned value will be ignored
        field_schema.update(
            slice={
                "description": "A slice object representing the set of indices specified by range(start, stop, step).",  # noqa: E501 # FIXME CoP
                "type": "object",
                "properties": {
                    "start": {
                        "description": "The starting index of the slice.",
                        "type": "integer",
                    },
                    "stop": {
                        "description": "The stopping index of the slice.",
                        "type": "integer",
                    },
                    "step": {
                        "description": "The number of steps between indices.",
                        "type": "integer",
                    },
                },
            }
        )

    @classmethod
    def validate(cls, v):
        if not isinstance(v, slice):
            raise TypeError("slice required")  # noqa: TRY003 # FIXME CoP
        return cls(v)


BatchSlice: TypeAlias = Union[
    Sequence[Union[StrictInt, None]], SliceValidator, StrictInt, StrictStr
]


def build_batch_filter(  # noqa: C901 #  too complex
    data_connector_query_dict: Optional[
        Dict[
            str,
            Optional[
                Union[
                    int,
                    list,
                    tuple,
                    Union[slice, SliceValidator],
                    str,
                    Union[Dict, IDDict],
                    Callable,
                ]
            ],
        ]
    ] = None,
):
    if not data_connector_query_dict:
        return BatchFilter(
            custom_filter_function=None,
            batch_filter_parameters=None,
            index=None,
            limit=None,
        )
    data_connector_query_keys: set = set(data_connector_query_dict.keys())
    if not data_connector_query_keys <= BatchFilter.RECOGNIZED_KEYS:
        raise gx_exceptions.BatchFilterError(  # noqa: TRY003 # FIXME CoP
            f"""Unrecognized data_connector_query key(s):
"{data_connector_query_keys - BatchFilter.RECOGNIZED_KEYS!s}" detected.
            """
        )
    custom_filter_function: Optional[Callable] = data_connector_query_dict.get(  # type: ignore[assignment] # FIXME CoP
        "custom_filter_function"
    )
    if custom_filter_function and not isinstance(custom_filter_function, Callable):  # type: ignore[arg-type] # FIXME CoP
        raise gx_exceptions.BatchFilterError(  # noqa: TRY003 # FIXME CoP
            f"""The type of a custom_filter must be a function (Python "Callable").  The type given is
"{type(custom_filter_function)!s}", which is illegal.
            """  # noqa: E501 # FIXME CoP
        )
    batch_filter_parameters: Optional[Union[dict, IDDict]] = data_connector_query_dict.get(  # type: ignore[assignment] # FIXME CoP
        "batch_filter_parameters"
    )
    if batch_filter_parameters:
        if not isinstance(batch_filter_parameters, dict):
            raise gx_exceptions.BatchFilterError(  # noqa: TRY003 # FIXME CoP
                f"""The type of batch_filter_parameters must be a dictionary (Python "dict").  The type given is
"{type(batch_filter_parameters)!s}", which is illegal.
                """  # noqa: E501 # FIXME CoP
            )
        if not all(isinstance(key, str) for key in batch_filter_parameters):
            raise gx_exceptions.BatchFilterError(  # noqa: TRY003 # FIXME CoP
                'All batch_filter_parameters keys must strings (Python "str").'
            )
        batch_filter_parameters = IDDict(batch_filter_parameters)
    index: Optional[BatchSlice] = data_connector_query_dict.get(  # type: ignore[assignment] # FIXME CoP
        "index"
    )
    limit: Optional[int] = data_connector_query_dict.get("limit")  # type: ignore[assignment] # FIXME CoP
    if limit and (not isinstance(limit, int) or limit < 0):
        raise gx_exceptions.BatchFilterError(  # noqa: TRY003 # FIXME CoP
            f"""The type of a limit must be an integer (Python "int") that is greater than or equal to 0.  The
type and value given are "{type(limit)!s}" and "{limit}", respectively, which is illegal.
            """  # noqa: E501 # FIXME CoP
        )
    if index is not None and limit is not None:
        raise gx_exceptions.BatchFilterError(  # noqa: TRY003 # FIXME CoP
            "Only one of index or limit, but not both, can be specified (specifying both is illegal)."  # noqa: E501 # FIXME CoP
        )
    parsed_index: slice | None = parse_batch_slice(batch_slice=index) if index is not None else None
    return BatchFilter(
        custom_filter_function=custom_filter_function,
        batch_filter_parameters=batch_filter_parameters,  # type: ignore[arg-type] # FIXME CoP
        index=parsed_index,
        limit=limit,
    )


def _batch_slice_string_to_slice_params(batch_slice: str) -> list[int | None]:
    # trim whitespace
    parsed_batch_slice = batch_slice.strip()

    slice_params: list[int | None] = []
    if parsed_batch_slice:
        # determine if bracket or slice() notation and choose delimiter
        delimiter: str = ":"
        if (parsed_batch_slice[0] in "[(") and (parsed_batch_slice[-1] in ")]"):
            parsed_batch_slice = parsed_batch_slice[1:-1]
        elif parsed_batch_slice.startswith("slice(") and parsed_batch_slice.endswith(")"):
            parsed_batch_slice = parsed_batch_slice[6:-1]
            delimiter = ","

        # split and convert string to int
        for param in parsed_batch_slice.split(delimiter):
            param = param.strip()  # noqa: PLW2901 # FIXME CoP
            if param and param != "None":
                try:
                    slice_params.append(int(param))
                except ValueError as e:
                    raise ValueError(  # noqa: TRY003 # FIXME CoP
                        f'Attempt to convert string slice index "{param}" to integer failed with message: {e}'  # noqa: E501 # FIXME CoP
                    )
            else:
                slice_params.append(None)

    return slice_params


def _batch_slice_from_string(batch_slice: str) -> slice:
    slice_params: list[int | None] = _batch_slice_string_to_slice_params(batch_slice=batch_slice)

    if len(slice_params) == 0:
        return slice(0, None, None)
    elif len(slice_params) == 1 and slice_params[0] is not None:
        return _batch_slice_from_int(batch_slice=slice_params[0])
    elif len(slice_params) == 2:  # noqa: PLR2004 # FIXME CoP
        return slice(slice_params[0], slice_params[1], None)
    elif len(slice_params) == 3:  # noqa: PLR2004 # FIXME CoP
        return slice(slice_params[0], slice_params[1], slice_params[2])
    else:
        raise ValueError(  # noqa: TRY003 # FIXME CoP
            f"batch_slice string must take the form of a python slice, but {batch_slice} was provided."  # noqa: E501 # FIXME CoP
        )


def _batch_slice_from_list_or_tuple(batch_slice: list[int] | tuple[int, ...]) -> slice:
    if len(batch_slice) == 0:
        return slice(0, None, None)
    elif len(batch_slice) == 1 and batch_slice[0] is not None:
        return slice(batch_slice[0] - 1, batch_slice[0])
    elif len(batch_slice) == 2:  # noqa: PLR2004 # FIXME CoP
        return slice(batch_slice[0], batch_slice[1])
    elif len(batch_slice) == 3:  # noqa: PLR2004 # FIXME CoP
        return slice(batch_slice[0], batch_slice[1], batch_slice[2])
    else:
        raise ValueError(  # noqa: TRY003 # FIXME CoP
            f'batch_slice sequence must be of length 0-3, but "{batch_slice}" was provided.'
        )


def _batch_slice_from_int(batch_slice: int) -> slice:
    if batch_slice == -1:
        return slice(batch_slice, None, None)
    else:
        return slice(batch_slice, batch_slice + 1, None)


def parse_batch_slice(batch_slice: Optional[BatchSlice]) -> slice:
    return_slice: slice
    if batch_slice is None:
        return_slice = slice(0, None, None)
    elif isinstance(batch_slice, slice):
        return_slice = batch_slice
    elif isinstance(batch_slice, SliceValidator):
        return_slice = slice(batch_slice.start, batch_slice.stop, batch_slice.step)
    elif isinstance(batch_slice, int) and not isinstance(batch_slice, bool):
        return_slice = _batch_slice_from_int(batch_slice=batch_slice)
    elif isinstance(batch_slice, str):
        return_slice = _batch_slice_from_string(batch_slice=batch_slice)
    elif isinstance(batch_slice, (list, tuple)):
        return_slice = _batch_slice_from_list_or_tuple(batch_slice=batch_slice)
    else:
        raise TypeError(  # noqa: TRY003 # FIXME CoP
            f"`batch_slice` should be of type `BatchSlice`, but type: {type(batch_slice)} was passed."  # noqa: E501 # FIXME CoP
        )
    logger.info(f"batch_slice: {batch_slice} was parsed to: {return_slice}")
    return return_slice


class BatchFilter:
    RECOGNIZED_KEYS: set = {
        "custom_filter_function",
        "batch_filter_parameters",
        "index",
        "limit",
    }

    def __init__(
        self,
        custom_filter_function: Optional[Callable] = None,
        batch_filter_parameters: Optional[IDDict] = None,
        index: Optional[Union[int, slice]] = None,
        limit: Optional[int] = None,
    ) -> None:
        self._custom_filter_function = custom_filter_function
        self._batch_filter_parameters = batch_filter_parameters
        self._index = index
        self._limit = limit

    @property
    def custom_filter_function(self) -> Optional[Callable]:
        return self._custom_filter_function

    @property
    def batch_filter_parameters(self) -> Optional[IDDict]:
        return self._batch_filter_parameters

    @property
    def index(self) -> Optional[Union[int, slice]]:
        return self._index

    @property
    def limit(self) -> int:
        return self._limit  # type: ignore[return-value] # FIXME CoP

    @override
    def __repr__(self) -> str:
        doc_fields_dict: dict = {
            "custom_filter_function": self._custom_filter_function,
            "batch_filter_parameters": self.batch_filter_parameters,
            "index": self.index,
            "limit": self.limit,
        }
        return str(doc_fields_dict)

    def select_from_data_connector_query(
        self, batch_definition_list: Optional[List[LegacyBatchDefinition]] = None
    ) -> List[LegacyBatchDefinition]:
        if batch_definition_list is None:
            return []
        filter_function: Callable
        if self.custom_filter_function:
            filter_function = self.custom_filter_function
        else:
            filter_function = self.best_effort_batch_definition_matcher()
        selected_batch_definitions: List[LegacyBatchDefinition]
        selected_batch_definitions = list(
            filter(
                lambda batch_definition: filter_function(
                    batch_identifiers=batch_definition.batch_identifiers,
                ),
                batch_definition_list,
            )
        )
        if len(selected_batch_definitions) == 0:
            return selected_batch_definitions

        if self.index is None:
            selected_batch_definitions = selected_batch_definitions[: self.limit]
        else:  # noqa: PLR5501 # FIXME CoP
            if isinstance(self.index, int):
                selected_batch_definitions = [selected_batch_definitions[self.index]]
            else:
                selected_batch_definitions = list(
                    itertools.chain.from_iterable([selected_batch_definitions[self.index]])
                )
        return selected_batch_definitions

    def best_effort_batch_definition_matcher(self) -> Callable:
        def match_batch_identifiers_to_batch_filter_params(
            batch_identifiers: dict,
        ) -> bool:
            if self.batch_filter_parameters:
                if not batch_identifiers:
                    return False

                for batch_filter_parameter, val in self.batch_filter_parameters.items():
                    if not (
                        batch_filter_parameter in batch_identifiers
                        and batch_identifiers[batch_filter_parameter] == val
                    ):
                        return False

            return True

        return match_batch_identifiers_to_batch_filter_params



import pytest

import great_expectations.exceptions as gx_exceptions
import great_expectations.expectations as gxe
from great_expectations.compatibility import sqlalchemy
from great_expectations.execution_engine.sqlite_execution_engine import SqliteExecutionEngine
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.expectations.registry import (
    get_expectation_impl,
    get_metric_provider,
    get_sqlalchemy_metric_provider,
)

# module level markers
pytestmark = pytest.mark.unit


def test_registry_basics():
    expectation = get_expectation_impl("expect_column_values_to_be_in_set")
    assert expectation == gxe.ExpectColumnValuesToBeInSet


def test_registry_from_configuration():
    configuration = ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "PClass", "value_set": [1, 2, 3]},
    )
    assert configuration._get_expectation_impl() == gxe.ExpectColumnValuesToBeInSet


def test_registry_raises_error_when_invalid_expectation_requested():
    with pytest.raises(gx_exceptions.ExpectationNotFoundError):
        get_expectation_impl("expect_something_in_beta")


def test_get_metric_provider_for_sqlalchemy_engine_subclass():
    # This is a regression test to show that we can register a metric for
    # a sqlalchemy engine subclass and it doesn't overwrite the default metric.
    metric_name = "column.standard_deviation.aggregate_fn"

    sa_metric_provider, sa_metric_provider_fn = get_sqlalchemy_metric_provider(
        metric_name,
    )

    # get the sqlite metric provider
    sqlite_metric_provider, sqlite_metric_provider_fn = get_metric_provider(
        metric_name,
        SqliteExecutionEngine(engine=sqlalchemy.create_engine("sqlite://")),
    )
    assert sa_metric_provider is not None
    assert sqlite_metric_provider is not None
    assert sa_metric_provider != sqlite_metric_provider
    assert sa_metric_provider_fn is not None
    assert sqlite_metric_provider_fn is not None
    assert sa_metric_provider_fn != sqlite_metric_provider_fn
